# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any

import json
import logging
import uuid
from datetime import datetime
from datetime import timedelta
from datetime import UTC

import sqlalchemy as sa
from nbxmpp.protocol import JID
from nbxmpp.structs import CommonError
from sqlalchemy.exc import IntegrityError
from sqlalchemy.exc import OperationalError
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import mapped_column
from sqlalchemy.orm import MappedAsDataclass
from sqlalchemy.orm import relationship

from gajim.common import app
from gajim.common.const import Trust
from gajim.common.events import DBMigrationFinished
from gajim.common.events import DBMigrationProgress
from gajim.common.events import DBMigrationStart
from gajim.common.storage.archive import models as mod
from gajim.common.storage.archive.const import ChatDirection
from gajim.common.storage.archive.const import MessageState
from gajim.common.storage.archive.const import MessageType
from gajim.common.storage.base import AlchemyStorage
from gajim.common.storage.base import is_unique_constraint_error

log = logging.getLogger("gajim.c.storage.archive.migration")


class MigrationBase(DeclarativeBase):
    pass


class LastArchiveMessage(MappedAsDataclass, MigrationBase, kw_only=True):
    __tablename__ = "last_archive_message"

    jid_id: Mapped[int] = mapped_column(sa.ForeignKey("jids.jid_id"), primary_key=True)
    remote: Mapped[Jids | None] = relationship(lazy="joined", viewonly=True, init=False)
    last_mam_id: Mapped[str | None]
    oldest_mam_timestamp: Mapped[float | None]
    last_muc_timestamp: Mapped[float | None]


class Jids(MappedAsDataclass, MigrationBase, kw_only=True):
    __tablename__ = "jids"

    jid_id: Mapped[int] = mapped_column(primary_key=True)
    jid: Mapped[str]
    type: Mapped[int]


class Logs(MappedAsDataclass, MigrationBase, kw_only=True):
    __tablename__ = "logs"

    log_line_id: Mapped[int] = mapped_column(primary_key=True)
    account_id: Mapped[int] = mapped_column(sa.ForeignKey("jids.jid_id"))
    account: Mapped[Jids | None] = relationship(
        lazy="joined", foreign_keys=[account_id], viewonly=True, init=False
    )

    jid_id: Mapped[int] = mapped_column(sa.ForeignKey("jids.jid_id"))
    remote: Mapped[Jids | None] = relationship(
        lazy="joined", foreign_keys=[jid_id], viewonly=True, init=False
    )

    contact_name: Mapped[str | None]
    occupant_id: Mapped[str | None]
    real_jid: Mapped[str | None]
    time: Mapped[float | None]
    kind: Mapped[int]
    show: Mapped[int | None]
    message: Mapped[str]
    error: Mapped[str | None]
    subject: Mapped[str | None]
    additional_data: Mapped[str | None]
    stanza_id: Mapped[str | None]
    message_id: Mapped[str | None]
    encryption: Mapped[str | None]
    encryption_state: Mapped[str | None]
    marker: Mapped[int | None]


KIND_MAPPING = {
    2: (MessageType.GROUPCHAT, ChatDirection.INCOMING),
    4: (MessageType.CHAT, ChatDirection.INCOMING),
    6: (MessageType.CHAT, ChatDirection.OUTGOING),
}


class Migration:
    def __init__(self, archive: AlchemyStorage, user_version: int) -> None:
        self._archive = archive
        self._engine = archive.get_engine()
        self._start_version = user_version

        self._account_pks: dict[JID, int] = {}
        self._remote_pks: dict[JID, int] = {}
        self._encryption_pks: dict[tuple[str | Trust, ...], int] = {}

        self._accounts: dict[str, str] = {}

        if user_version < 7:
            self._pre_v7(user_version)
        if user_version < 8:
            self._v8()
        if user_version < 9:
            self._v9()
        if user_version < 10:
            self._v10()
        if user_version < 12:
            self._v11_and_v12()
        if user_version < 13:
            self._v13()
        if user_version < 14:
            self._v14()
        if user_version < 15:
            self._v15()
        if user_version < 16:
            self._v16()
        if user_version < 17:
            self._v17()

        app.ged.raise_event(DBMigrationFinished())

    def _execute_multiple(self, statements: list[str]) -> None:
        with self._engine.begin() as conn:
            for stmt in statements:
                conn.execute(sa.text(stmt))

    def _execute_multiple_with_error(
        self, statements: list[tuple[str, str] | tuple[str, None]]
    ) -> None:
        with self._engine.begin() as conn:
            for stmt, error_string in statements:
                try:
                    conn.execute(sa.text(stmt))
                except OperationalError as e:
                    if error_string is not None:
                        assert e.orig is not None
                        sql_error = e.orig.args[0]
                        if error_string in sql_error:
                            log.warning('Ignore "%s" for %s', sql_error, e.statement)
                            continue
                    raise

    def _pre_v7(self, user_version: int) -> None:
        app.ged.raise_event(DBMigrationStart(version=7))

        if user_version == 0:
            # All migrations from 0.16.9 until 1.0.0
            statements = [
                'ALTER TABLE logs ADD COLUMN "account_id" INTEGER',
                'ALTER TABLE logs ADD COLUMN "stanza_id" TEXT',
                'ALTER TABLE logs ADD COLUMN "encryption" TEXT',
                'ALTER TABLE logs ADD COLUMN "encryption_state" TEXT',
                'ALTER TABLE logs ADD COLUMN "marker" INTEGER',
                'ALTER TABLE logs ADD COLUMN "additional_data" TEXT',
                """CREATE TABLE IF NOT EXISTS last_archive_message(
                    jid_id INTEGER PRIMARY KEY UNIQUE,
                    last_mam_id TEXT,
                    oldest_mam_timestamp TEXT,
                    last_muc_timestamp TEXT
                    )""",
                "PRAGMA user_version=1",
            ]

            self._execute_multiple(statements)

        if user_version < 2:
            statements = [
                (
                    "ALTER TABLE last_archive_message "
                    'ADD COLUMN "sync_threshold" INTEGER'
                ),
                "PRAGMA user_version=2",
            ]
            self._execute_multiple(statements)

        if user_version < 3:
            statements = [
                'ALTER TABLE logs ADD COLUMN "message_id" TEXT',
                "PRAGMA user_version=3",
            ]
            self._execute_multiple(statements)

        if user_version < 4:
            statements = [
                'ALTER TABLE logs ADD COLUMN "error" TEXT',
                "PRAGMA user_version=4",
            ]
            self._execute_multiple(statements)

        if user_version < 7:
            statements = [
                'ALTER TABLE logs ADD COLUMN "real_jid" TEXT',
                'ALTER TABLE logs ADD COLUMN "occupant_id" TEXT',
                "PRAGMA user_version=7",
            ]
            self._execute_multiple(statements)

    def _v8(self) -> None:
        app.ged.raise_event(DBMigrationStart(version=8))

        for account_name in app.settings.get_accounts():
            jid = app.get_jid_from_account(account_name)
            self._accounts[jid] = account_name

        mod.Base.metadata.create_all(self._engine)

        count_stmt = sa.select(sa.func.count(Logs.log_line_id)).where(
            Logs.kind.in_([2, 4, 6])
        )
        stmt = (
            sa.select(Logs)
            .where(Logs.kind.in_([2, 4, 6]))
            .execution_options(yield_per=1000)
        )

        with self._engine.begin() as conn:
            count = conn.execute(count_stmt).scalar()
            assert count is not None
            for i, log_row in enumerate(self._archive.get_session().scalars(stmt)):
                self._process_message_row(conn, log_row)

                if i % 1000 == 0:
                    app.ged.raise_event(DBMigrationProgress(count=count, progress=i))

            app.ged.raise_event(DBMigrationProgress(count=count, progress=count))

            stmt = sa.select(LastArchiveMessage)

            account_pks = self._get_account_pks(conn)

            for archive_row in self._archive.get_session().scalars(stmt):
                self._process_archive_row(conn, archive_row, account_pks)

            conn.execute(sa.text("DROP TABLE last_archive_message"))
            conn.execute(sa.text("DROP TABLE logs"))
            conn.execute(sa.text("DROP TABLE jids"))
            conn.execute(sa.text("DROP TABLE IF EXISTS unread_messages"))
            conn.execute(sa.text("PRAGMA user_version=8"))

    def _v9(self) -> None:
        app.ged.raise_event(DBMigrationStart(version=9))
        statements = [
            "CREATE INDEX IF NOT EXISTS idx_stanza_id ON message(stanza_id, fk_remote_pk, fk_account_pk);",
            "PRAGMA user_version=9",
        ]

        self._execute_multiple(statements)

    def _v10(self) -> None:
        app.ged.raise_event(DBMigrationStart(version=10))
        self._execute_multiple(["PRAGMA foreign_keys=OFF"])

        with self._engine.begin() as conn:
            stmt = sa.select(mod.Message.fk_occupant_pk).union(
                sa.select(mod.Moderation.fk_occupant_pk),
                sa.select(mod.DisplayedMarker.fk_occupant_pk),
            )
            occupant_pks: set[int | None] = set(conn.scalars(stmt))
            occupant_pks.discard(None)

            conn.execute(
                sa.delete(mod.Occupant).where(mod.Occupant.pk.not_in(occupant_pks))
            )
            conn.commit()

        app.ged.raise_event(DBMigrationProgress(count=100, progress=50))

        with self._engine.begin() as conn:
            stmt = sa.select(mod.Occupant.fk_remote_pk).union(
                sa.select(mod.Occupant.fk_real_remote_pk),
                sa.select(mod.SecurityLabel.fk_remote_pk),
                sa.select(mod.MessageError.fk_remote_pk),
                sa.select(mod.Moderation.fk_remote_pk),
                sa.select(mod.DisplayedMarker.fk_remote_pk),
                sa.select(mod.Receipt.fk_remote_pk),
                sa.select(mod.MAMArchiveState.fk_remote_pk),
                sa.select(mod.Message.fk_remote_pk),
            )

            occupant_pks: set[int | None] = set(conn.scalars(stmt))
            occupant_pks.discard(None)

            conn.execute(
                sa.delete(mod.Remote).where(mod.Remote.pk.not_in(occupant_pks))
            )

        self._execute_multiple(["PRAGMA foreign_keys=ON", "PRAGMA user_version=10"])

    def _v11_and_v12(self) -> None:
        app.ged.raise_event(DBMigrationStart(version=12))
        mod.Base.metadata.create_all(self._engine)
        self._execute_multiple(["PRAGMA user_version=12"])

    def _v13(self) -> None:
        app.ged.raise_event(DBMigrationStart(version=13))
        statements = [
            (
                'ALTER TABLE occupant ADD COLUMN "blocked" INTEGER NOT NULL DEFAULT (0)',
                "duplicate column",
            ),
            ("PRAGMA user_version=13", None),
        ]
        self._execute_multiple_with_error(statements)

    def _v14(self) -> None:
        app.ged.raise_event(DBMigrationStart(version=14))
        statements = [
            (
                "CREATE INDEX IF NOT EXISTS idx_displayed_marker_last ON displayed_marker "
                "(fk_remote_pk, fk_account_pk, timestamp DESC)"
            ),
            (
                "CREATE INDEX IF NOT EXISTS idx_displayed_marker_last_gc ON displayed_marker "
                "(fk_remote_pk, fk_account_pk, fk_occupant_pk, timestamp DESC)"
            ),
            "PRAGMA user_version=14",
        ]
        self._execute_multiple(statements)

    def _v15(self) -> None:
        # In the past some PM messages were not correctly discovered
        # Find them and set the type and resource
        app.ged.raise_event(DBMigrationStart(version=15))
        unique_remotes_stmt = (
            sa.select(mod.Message.fk_remote_pk).distinct().where(mod.Message.type == 3)
        )
        remotes_stmt = sa.select(mod.Remote).where(
            mod.Remote.pk.in_(unique_remotes_stmt.scalar_subquery())
        )

        app.ged.raise_event(DBMigrationProgress(count=100, progress=10))
        with self._archive.get_session() as s:
            remotes = s.execute(remotes_stmt).scalars().all()
            for remote in remotes:
                stmt = (
                    sa.update(mod.Message)
                    .where(mod.Message.fk_remote_pk == remote.pk, mod.Message.type != 3)
                    .values(type=3, resource=remote.jid.resource)
                )
                s.execute(stmt)
            s.commit()

        self._execute_multiple(["PRAGMA user_version=15"])

    def _v16(self) -> None:
        app.ged.raise_event(DBMigrationStart(version=16))
        mod.Base.metadata.create_all(self._engine)
        self._execute_multiple(["PRAGMA user_version=16"])

    def _v17(self) -> None:
        app.ged.raise_event(DBMigrationStart(version=17))
        if self._start_version == 16:
            # if the start version is < 16, the version 16 migration will
            # create the current table, no need to alter it
            self._execute_multiple(
                [
                    'ALTER TABLE opengraph DROP COLUMN "image"',
                    'ALTER TABLE opengraph DROP COLUMN "site_name"',
                    'ALTER TABLE opengraph DROP COLUMN "type"',
                    'ALTER TABLE opengraph RENAME COLUMN "url" TO "about"',
                    'ALTER TABLE opengraph ADD COLUMN "image_type" TEXT',
                    'ALTER TABLE opengraph ADD COLUMN "image_bytes" BLOB',
                ]
            )
        self._execute_multiple(["PRAGMA user_version=17"])

    def _get_account_pks(self, conn: sa.Connection) -> list[int]:
        account_pks: list[int] = []
        for account in app.settings.get_accounts():
            jid_str = app.get_jid_from_account(account)
            try:
                jid = JID.from_string(jid_str)
            except Exception:
                log.warning("Unable to parse account: %s", jid_str)
                continue

            pk = self._get_account_pk(conn, jid)
            account_pks.append(pk)

        return account_pks

    def _process_archive_row(
        self,
        conn: sa.Connection,
        archive_row: LastArchiveMessage,
        account_pks: list[int],
    ) -> None:
        if archive_row.remote is None:
            log.warning(
                "Unable to migrate mam state because jid_id %s was not found",
                archive_row.jid_id,
            )
            return

        try:
            remote_jid = JID.from_string(archive_row.remote.jid)
        except Exception as error:
            log.warning("Invalid JID found, unable to migrate archive state: %s", error)
            return

        remote_pk = self._get_remote_pk(conn, remote_jid)

        to_stanza_id = archive_row.last_mam_id
        to_stanza_ts = archive_row.last_muc_timestamp
        if archive_row.last_mam_id is None or archive_row.last_muc_timestamp is None:
            to_stanza_id = None
            to_stanza_ts = None

        if to_stanza_ts is not None:
            to_stanza_ts = datetime.fromtimestamp(float(to_stanza_ts), tz=UTC)

        from_stanza_ts = None
        if archive_row.oldest_mam_timestamp is not None:
            from_stanza_ts = datetime.fromtimestamp(
                float(archive_row.oldest_mam_timestamp), tz=UTC
            )

        if (from_stanza_ts, to_stanza_id, to_stanza_ts) == (None, None, None):
            return

        for account_pk in account_pks:
            try:
                conn.execute(
                    sa.insert(mod.MAMArchiveState).values(
                        fk_account_pk=account_pk,
                        fk_remote_pk=remote_pk,
                        from_stanza_ts=from_stanza_ts,
                        to_stanza_id=to_stanza_id,
                        to_stanza_ts=to_stanza_ts,
                    )
                )
            except IntegrityError:
                log.warning(
                    "Unable to migrate mam archive state, because it was already migrated"
                )
                return

    def _process_message_row(self, conn: sa.Connection, log_row: Logs) -> None:
        m_type, direction = KIND_MAPPING[log_row.kind]

        if log_row.time is None:
            log.warning("Unable to migrate message because timestamp is empty")
            return

        if log_row.account is None:
            log.warning(
                "Unable to migrate message because account_id %s was not found",
                log_row.account_id,
            )
            return

        if log_row.remote is None:
            log.warning(
                "Unable to migrate message because jid_id %s was not found",
                log_row.jid_id,
            )
            return

        try:
            account_jid = JID.from_string(log_row.account.jid)
        except Exception:
            log.warning(
                "Unable to migrate message because of invalid account jid: %s",
                log_row.account.jid,
            )
            return

        try:
            remote_jid = JID.from_string(log_row.remote.jid)
        except Exception:
            log.warning(
                "Unable to migrate message because of invalid remote jid: %s",
                log_row.remote.jid,
            )
            return

        account_pk = self._get_account_pk(conn, account_jid)
        remote_pk = self._get_remote_pk(conn, remote_jid)

        try:
            additional_data = json.loads(log_row.additional_data or "{}")
        except Exception as error:
            raise ValueError(
                f"failed to parse additional_data: {error} - {log_row.additional_data}"
            )

        user_timestamp = additional_data.get("user_timestamp")
        user_delay_ts = None
        if user_timestamp is not None:
            user_delay_ts = datetime.fromtimestamp(user_timestamp, UTC)

        timestamp = datetime.fromtimestamp(log_row.time, UTC)

        encryption_pk = self._insert_encryption_data(conn, additional_data)

        text = log_row.message
        corrected = additional_data.get("corrected")
        if corrected is not None:
            text = corrected.get("original_text") or text

        message_data: dict[str, Any] = {
            "fk_account_pk": account_pk,
            "fk_remote_pk": remote_pk,
            "resource": log_row.contact_name,
            "type": m_type,
            "direction": direction,
            "timestamp": timestamp,
            "state": MessageState.ACKNOWLEDGED,
            "id": log_row.message_id,
            "stanza_id": log_row.stanza_id,
            "text": text,
            "user_delay_ts": user_delay_ts,
            "fk_encryption_pk": encryption_pk,
        }

        try:
            pk = conn.execute(
                sa.insert(mod.Message).returning(mod.Message.pk), [message_data]
            ).scalar()
        except IntegrityError as error:
            if "message.stanza_id" not in error.args[0]:
                raise

            message_data["stanza_id"] = str(uuid.uuid4())

            pk = conn.execute(
                sa.insert(mod.Message).returning(mod.Message.pk), [message_data]
            ).scalar()

        assert pk is not None

        self._insert_oob_data(conn, pk, additional_data)
        self._insert_error_data(conn, account_pk, remote_pk, log_row, timestamp)

    def _get_account_pk(self, conn: sa.Connection, jid: JID) -> int:
        pk = self._account_pks.get(jid)
        if pk is not None:
            return pk

        pk = conn.execute(
            sa.insert(mod.Account).values(jid=jid).returning(mod.Account.pk)
        ).scalar()

        if pk is None:
            raise ValueError("Failed to insert account")

        self._account_pks[jid] = pk
        return pk

    def _get_remote_pk(self, conn: sa.Connection, jid: JID) -> int:
        pk = self._remote_pks.get(jid)
        if pk is not None:
            return pk

        pk = conn.execute(
            sa.insert(mod.Remote).values(jid=jid).returning(mod.Remote.pk)
        ).scalar()

        if pk is None:
            raise ValueError("Failed to insert remote")

        self._remote_pks[jid] = pk
        return pk

    def _insert_oob_data(
        self,
        conn: sa.Connection,
        fk_message_pk: int,
        additional_data: dict[str, Any],
    ) -> dict[str, Any] | None:
        if not additional_data:
            return None

        gajim_data = additional_data.get("gajim")
        if gajim_data is None:
            return None

        url = gajim_data.get("oob_url")
        description = gajim_data.get("oob_desc")

        if url is None:
            return None

        oob_data = {
            "fk_message_pk": fk_message_pk,
            "url": url,
            "description": description,
        }

        conn.execute(sa.insert(mod.OOB), [oob_data])

    def _insert_error_data(
        self,
        conn: sa.Connection,
        fk_account_pk: int,
        fk_remote_pk: int,
        log_row: Logs,
        timestamp: datetime,
    ) -> dict[str, Any] | None:
        if log_row.error is None:
            return None

        try:
            error = CommonError.from_string(log_row.error)
        except Exception:
            return None

        assert log_row.message_id is not None

        by = error.by
        e_type = error.type
        text = error.get_text() or None
        condition = error.condition
        condition_text = error.condition_data or None

        if e_type is None or condition is None:
            return None

        error_data = {
            "fk_account_pk": fk_account_pk,
            "fk_remote_pk": fk_remote_pk,
            "message_id": log_row.message_id,
            "by": by,
            "type": e_type,
            "text": text,
            "condition": condition,
            "condition_text": condition_text,
            "timestamp": timestamp + timedelta(seconds=1),
        }

        try:
            conn.execute(sa.insert(mod.MessageError), [error_data])
        except IntegrityError as error:
            if is_unique_constraint_error(error):
                return

    def _insert_encryption_data(
        self, conn: sa.Connection, additional_data: dict[str, Any]
    ) -> int | None:
        if not additional_data:
            return None

        encrypted = additional_data.get("encrypted")
        if encrypted is None:
            return None

        protocol = encrypted.get("name")
        key = encrypted.get("fingerprint")
        trust = encrypted.get("trust")

        if protocol not in ("OpenPGP", "OMEMO", "PGP"):
            return None

        if key is None:
            key = "Unknown"

        if trust is None:
            trust = Trust.VERIFIED

        else:
            try:
                trust = Trust(trust)
            except Exception:
                trust = Trust.UNTRUSTED

        encryption_data = {"protocol": protocol, "key": key, "trust": trust}
        encryption_pk = self._encryption_pks.get(tuple(encryption_data.values()))
        if encryption_pk is not None:
            return encryption_pk

        pk = conn.execute(
            sa.insert(mod.Encryption).returning(mod.Encryption.pk),
            [encryption_data],
        ).scalar()

        if pk is None:
            return None

        data_tp = tuple(encryption_data.values())
        self._encryption_pks[data_tp] = pk
        return pk


def run(archive: Any, user_version: int) -> None:
    Migration(archive, user_version)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any

import math
from io import BytesIO
from pathlib import Path

from gi.repository import GdkPixbuf
from gi.repository import GLib
from PIL import Image
from PIL import ImageFile

ImageFile.LOAD_TRUNCATED_IMAGES = True


def get_image_orientation(image: Image.Image) -> int:
    exif = image.getexif()
    exif_orientation_code = 274
    orientation = exif.get(exif_orientation_code, 1)
    degrees = {1: 0, 3: 180, 6: 270, 8: 90}.get(orientation, 0)

    return degrees


def create_thumbnail(
    input_: bytes | Path,
    output: Path | None,
    size: int,
    mime_type: str,
) -> tuple[bytes, dict[str, Any]]:
    if isinstance(input_, Path):
        data = input_.read_bytes()
    else:
        data = input_

    try:
        thumbnail_bytes, metadata = _create_thumbnail_with_pil(data, size)
    except (Image.DecompressionBombError, Image.DecompressionBombWarning):
        # Don't try to process image further
        raise

    except Exception:
        thumbnail_bytes, metadata = _create_thumbnail_with_pixbuf(data, size, mime_type)

    assert thumbnail_bytes is not None
    if output is not None:
        output.write_bytes(thumbnail_bytes)
    return thumbnail_bytes, metadata


def _create_thumbnail_with_pil(data: bytes, size: int) -> tuple[bytes, dict[str, Any]]:
    # Reads data and returns thumbnail bytes in PNG format

    metadata: dict[str, Any] = {}
    input_file = BytesIO(data)
    try:
        image = Image.open(input_file)
    except Exception:
        input_file.close()
        raise

    image_orientation = get_image_orientation(image)
    if image_orientation != 0:
        image = image.rotate(image_orientation, expand=True)

    image_width, image_height = image.size
    n_frames = getattr(image, "n_frames", 1)
    if size > image_width and size > image_height and n_frames == 1:
        image.close()
        input_file.close()
        return data, metadata

    output_file = BytesIO()

    image.thumbnail((size, size))
    image.save(output_file, format="png", optimize=True, save_all=False)

    bytes_ = output_file.getvalue()

    image.close()
    input_file.close()
    output_file.close()

    return bytes_, metadata


def _create_thumbnail_with_pixbuf(
    data: bytes, size: int, mime_type: str
) -> tuple[bytes, dict[str, Any]]:
    # Reads data and returns thumbnail bytes in PNG format

    metadata: dict[str, Any] = {}

    try:
        # Try to create GdKPixbuf loader with fixed mime-type to
        # fix mime-type detection for HEIF images on some systems
        loader = GdkPixbuf.PixbufLoader.new_with_mime_type(mime_type)
    except GLib.Error:
        loader = GdkPixbuf.PixbufLoader()

    loader.write(data)
    loader.close()

    pixbuf = loader.get_pixbuf()
    if pixbuf is None:
        raise ValueError("Loading pixbuf failed")

    if size > pixbuf.get_width() and size > pixbuf.get_height():
        return data, metadata

    width, height = get_thumbnail_size(pixbuf, size)
    thumbnail = pixbuf.scale_simple(width, height, GdkPixbuf.InterpType.BILINEAR)
    if thumbnail is None:
        raise ValueError("scale_simple() failed")

    _error, bytes_ = thumbnail.save_to_bufferv("png", [], [])

    return bytes_, metadata


def get_thumbnail_size(pixbuf: GdkPixbuf.Pixbuf, size: int) -> tuple[int, int]:
    # Calculates the new thumbnail size while preserving the aspect ratio
    image_width = pixbuf.get_width()
    image_height = pixbuf.get_height()

    if image_width > image_height:
        if image_width > size:
            image_height = math.ceil(size / float(image_width) * image_height)
            image_width = int(size)
    else:
        if image_height > size:
            image_width = math.ceil(size / float(image_height) * image_width)
            image_height = int(size)

    return image_width, image_height



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from gi.repository import Gdk
from gi.repository import GObject
from gi.repository import Gtk

from gajim.common import app
from gajim.common.i18n import _

from gajim.gtk.alert import ConfirmationAlertDialog
from gajim.gtk.dropdown import GajimDropDown
from gajim.gtk.util.misc import get_ui_string
from gajim.gtk.util.window import get_app_window
from gajim.gtk.window import GajimAppWindow


@Gtk.Template(string=get_ui_string("manage_proxies.ui"))
class ManageProxies(GajimAppWindow):
    __gtype_name__ = "ManageProxies"

    _proxies_treeview: Gtk.TreeView = Gtk.Template.Child()
    _treeview_selection1: Gtk.TreeSelection = Gtk.Template.Child()
    _add_proxy_button: Gtk.Button = Gtk.Template.Child()
    _remove_proxy_button: Gtk.Button = Gtk.Template.Child()
    _settings_grid: Gtk.Grid = Gtk.Template.Child()
    _proxypass_entry: Gtk.Entry = Gtk.Template.Child()
    _proxyuser_entry: Gtk.Entry = Gtk.Template.Child()
    _useauth_checkbutton: Gtk.CheckButton = Gtk.Template.Child()
    _proxyport_entry: Gtk.Entry = Gtk.Template.Child()
    _proxyhost_entry: Gtk.Entry = Gtk.Template.Child()
    _proxytype_dropdown: GajimDropDown[str] = Gtk.Template.Child()
    _proxyname_entry: Gtk.Entry = Gtk.Template.Child()

    def __init__(self, transient_for: Gtk.Window | None = None) -> None:
        GajimAppWindow.__init__(
            self,
            name="ManageProxies",
            title=_("Manage Proxies"),
            default_width=500,
            transient_for=transient_for,
            modal=True,
        )

        self._proxytype_dropdown.set_data(
            {
                "http": "HTTP",
                "socks5": "SOCKS5",
            }
        )

        self._connect(
            self._proxies_treeview,
            "cursor-changed",
            self._on_proxies_treeview_cursor_changed,
        )
        self._connect(
            self._add_proxy_button, "clicked", self._on_add_proxy_button_clicked
        )
        self._connect(
            self._remove_proxy_button,
            "clicked",
            self._on_remove_proxy_button_clicked,
        )

        self._connect(
            self._proxyuser_entry, "changed", self._on_proxyuser_entry_changed
        )
        self._connect(
            self._proxypass_entry, "changed", self._on_proxypass_entry_changed
        )
        self._connect(self._useauth_checkbutton, "toggled", self._on_useauth_toggled)
        self._connect(
            self._proxyport_entry, "changed", self._on_proxyport_entry_changed
        )
        self._connect(
            self._proxyhost_entry, "changed", self._on_proxyhost_entry_changed
        )
        self._connect(
            self._proxytype_dropdown,
            "notify::selected",
            self._on_proxytype_dropdown_selected,
        )
        self._connect(
            self._proxyname_entry, "changed", self._on_proxyname_entry_changed
        )

        self._init_list()
        self._block_signal = False

        controller = Gtk.EventControllerKey()
        self._connect(
            self.get_default_controller(),
            "key-pressed",
            self._on_proxies_treeview_key_pressed,
        )
        self._proxies_treeview.add_controller(controller)

    def _cleanup(self) -> None:
        window_pref = get_app_window("Preferences")
        window_account_wizard = get_app_window("AccountWizard")
        if window_pref is not None:
            window_pref.update_proxy_list()
        if window_account_wizard is not None:
            window_account_wizard.update_proxy_list()

    def _fill_proxies_treeview(self) -> None:
        model = self._proxies_treeview.get_model()
        assert isinstance(model, Gtk.ListStore)
        model.clear()
        for proxy in app.settings.get_proxies():
            iter_ = model.append()
            model.set_value(iter_, 0, proxy)

    def _init_list(self) -> None:
        self._remove_proxy_button.set_sensitive(False)
        self._settings_grid.set_sensitive(False)
        model = Gtk.ListStore(str)
        self._proxies_treeview.set_model(model)
        col = Gtk.TreeViewColumn(title=_("Proxies"))
        self._proxies_treeview.append_column(col)
        renderer = Gtk.CellRendererText()
        col.pack_start(renderer, True)
        col.add_attribute(renderer, "text", 0)
        self._fill_proxies_treeview()
        self._proxytype_dropdown.select_first()

    def _on_add_proxy_button_clicked(self, _button: Gtk.Button) -> None:
        model = self._proxies_treeview.get_model()
        assert isinstance(model, Gtk.ListStore)
        proxies = app.settings.get_proxies()
        i = 1
        while "proxy" + str(i) in proxies:
            i += 1

        proxy_name = "proxy" + str(i)
        app.settings.add_proxy(proxy_name)
        iter_ = model.append()
        model.set_value(iter_, 0, proxy_name)
        self._proxies_treeview.set_cursor(model.get_path(iter_))

    def _on_remove_proxy_button_clicked(self, _button: Gtk.Button) -> None:
        self._remove_selected_proxy()

    def _remove_selected_proxy(self) -> None:
        def _on_response() -> None:
            sel = self._proxies_treeview.get_selection()
            if not sel:
                return
            (model, iter_) = sel.get_selected()
            if not iter_:
                return

            assert isinstance(model, Gtk.ListStore)
            proxy = model[iter_][0]
            model.remove(iter_)
            app.settings.remove_proxy(proxy)
            self._remove_proxy_button.set_sensitive(False)
            self._block_signal = True
            self._on_proxies_treeview_cursor_changed(self._proxies_treeview)
            self._block_signal = False

        ConfirmationAlertDialog(
            _("Remove Proxy?"),
            _("Do you want to remove this proxy?"),
            confirm_label=_("_Remove"),
            appearance="destructive",
            callback=_on_response,
            parent=self,
        )

    def _on_useauth_toggled(self, checkbutton: Gtk.CheckButton) -> None:
        if self._block_signal:
            return
        act = checkbutton.get_active()
        proxy = self._proxyname_entry.get_text()
        app.settings.set_proxy_setting(proxy, "useauth", act)
        self._proxyuser_entry.set_sensitive(act)
        self._proxypass_entry.set_sensitive(act)

    def _on_proxies_treeview_cursor_changed(self, treeview: Gtk.TreeView) -> None:
        self._block_signal = True
        self._proxyhost_entry.set_text("")
        self._proxyport_entry.set_text("")
        self._proxyuser_entry.set_text("")
        self._proxypass_entry.set_text("")

        sel = treeview.get_selection()
        if not sel:
            self._proxyname_entry.set_text("")
            self._settings_grid.set_sensitive(False)
            self._block_signal = False
            return

        (model, iter_) = sel.get_selected()
        if iter_ is None:
            return

        proxy = model[iter_][0]
        self._proxyname_entry.set_text(proxy)

        self._remove_proxy_button.set_sensitive(True)
        self._proxyname_entry.set_editable(True)

        self._settings_grid.set_sensitive(True)

        settings = app.settings.get_proxy_settings(proxy)

        self._proxyhost_entry.set_text(str(settings["host"]))
        self._proxyport_entry.set_text(str(settings["port"]))
        self._proxyuser_entry.set_text(str(settings["user"]))
        self._proxypass_entry.set_text(str(settings["pass"]))

        self._proxytype_dropdown.select_key(str(settings["type"]))

        self._useauth_checkbutton.set_active(bool(settings["useauth"]))
        act = self._useauth_checkbutton.get_active()
        self._proxyuser_entry.set_sensitive(act)
        self._proxypass_entry.set_sensitive(act)

        self._block_signal = False

    def _on_proxies_treeview_key_pressed(
        self,
        _event_controller_key: Gtk.EventControllerKey,
        keyval: int,
        _keycode: int,
        _state: Gdk.ModifierType,
    ) -> bool:
        if keyval == Gdk.KEY_Delete:
            self._remove_selected_proxy()
            return Gdk.EVENT_STOP

        return Gdk.EVENT_PROPAGATE

    def _on_proxyname_entry_changed(self, entry: Gtk.Entry) -> None:
        if self._block_signal:
            return
        sel = self._proxies_treeview.get_selection()
        if not sel:
            return
        (model, iter_) = sel.get_selected()
        if not iter_:
            return
        old_name = model.get_value(iter_, 0)
        assert isinstance(old_name, str)
        new_name = entry.get_text()
        if new_name == "":
            return
        if new_name == old_name:
            return

        app.settings.rename_proxy(old_name, new_name)
        assert isinstance(model, Gtk.ListStore)
        model.set_value(iter_, 0, new_name)

    def _on_proxytype_dropdown_selected(
        self, dropdown: GajimDropDown[str], _param: GObject.ParamSpec
    ) -> None:
        if self._block_signal:
            return

        type_ = self._proxytype_dropdown.get_selected_key()
        self._proxyhost_entry.set_sensitive(True)
        self._proxyport_entry.set_sensitive(True)
        proxy = self._proxyname_entry.get_text()
        app.settings.set_proxy_setting(proxy, "type", type_)

    def _on_proxyhost_entry_changed(self, entry: Gtk.Entry) -> None:
        if self._block_signal:
            return
        value = entry.get_text()
        proxy = self._proxyname_entry.get_text()
        app.settings.set_proxy_setting(proxy, "host", value)

    def _on_proxyport_entry_changed(self, entry: Gtk.Entry) -> None:
        if self._block_signal:
            return
        value = entry.get_text()
        try:
            value = int(value)
        except Exception:
            value = 0
        proxy = self._proxyname_entry.get_text()
        app.settings.set_proxy_setting(proxy, "port", value)

    def _on_proxyuser_entry_changed(self, entry: Gtk.Entry) -> None:
        if self._block_signal:
            return
        value = entry.get_text()
        proxy = self._proxyname_entry.get_text()
        app.settings.set_proxy_setting(proxy, "user", value)

    def _on_proxypass_entry_changed(self, entry: Gtk.Entry) -> None:
        if self._block_signal:
            return
        value = entry.get_text()
        proxy = self._proxyname_entry.get_text()
        app.settings.set_proxy_setting(proxy, "pass", value)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import typing

import logging

from gi.repository import Gdk

from gajim.common import app

try:
    from gi.repository import Gst
except Exception:
    if typing.TYPE_CHECKING:
        from gi.repository import Gst

log = logging.getLogger("gajim.gtk.gstreamer")


def create_video_elements() -> tuple[Gst.Element, Gdk.Paintable, str] | None:
    if not app.is_installed("GST"):
        return None

    gtksink = Gst.ElementFactory.make("gtk4paintablesink", None)
    if gtksink is None:
        return None

    paintable = gtksink.get_property("paintable")
    if paintable.props.gl_context is not None:
        sink = Gst.ElementFactory.make("glsinkbin", None)
        if sink is None:
            return None

        log.info("Using GL")
        sink.set_property("sink", gtksink)
        name = "gtkglsink"

    else:
        sink = Gst.Bin.new()
        convert = Gst.ElementFactory.make("videoconvert", None)
        if convert is None:
            return None

        sink.add(convert)
        sink.add(gtksink)
        convert.link(gtksink)

        pad = convert.get_static_pad("sink")
        if pad is None:
            return None

        log.info("Not using GL")
        sink.add_pad(Gst.GhostPad.new("sink", pad))

        name = "gtksink"

    return sink, paintable, name



# Copyright (C) 2009-2014 Yann Leboulanger <asterix AT lagaule.org>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import typing
from typing import Any

import logging
from datetime import datetime
from datetime import UTC

from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject
from nbxmpp.structs import LocationData

from gajim.common import app
from gajim.common.events import LocationChanged

if app.is_installed("GEOCLUE") or typing.TYPE_CHECKING:
    import gi

    gi.require_version("Geoclue", "2.0")
    from gi.repository import Geoclue

log = logging.getLogger("gajim.c.dbus.location")


class LocationListener:
    _instance: LocationListener | None = None

    @classmethod
    def get(cls) -> LocationListener:
        if cls._instance is None:
            cls._instance = cls()
        if app.is_installed("GEOCLUE") and not cls._instance._running:
            cls._instance.start()
        return cls._instance

    def __init__(self) -> None:
        self._data: dict[str, Any] = {}
        self.location_info: dict[str, Any] = {}
        self.simple: Geoclue.Simple | None = None
        self._current_location: LocationData | None = None
        self._running = False

    def _emit(self, info: LocationData | None) -> None:
        self._current_location = info
        app.ged.raise_event(LocationChanged(info=info))

    @property
    def current_location(self) -> LocationData | None:
        return self._current_location

    def _on_location_update(self, simple: Geoclue.Simple, *args: Any) -> None:
        location = simple.get_location()
        timestamp = location.get_property("timestamp")[0]
        lat = location.get_property("latitude")
        lon = location.get_property("longitude")
        alt = location.get_property("altitude")
        # in XEP-0080 it's horizontal accuracy
        acc = location.get_property("accuracy")

        # update data with info we just received
        self._data = {"lat": lat, "lon": lon, "alt": alt, "accuracy": acc}
        self._data["timestamp"] = self._timestamp_to_string(timestamp)
        self._send_location()

    def _on_client_update(self, client: Geoclue.Client, *args: Any) -> None:
        if not client.get_property("active"):
            self._emit(None)

    def _on_simple_ready(self, _obj: GObject.Object, result: Gio.AsyncResult) -> None:
        try:
            self.simple = Geoclue.Simple.new_finish(result)
        except GLib.Error as error:
            log.warning("Could not enable geolocation: %s", error.message)
            self._running = False
        else:
            assert self.simple is not None
            self.simple.connect("notify::location", self._on_location_update)
            client = self.simple.get_client()
            # Inside the flatpak sandbox client will be None,
            # because the location portal is used instead of Geoclue directly.
            if client is not None:
                client.connect("notify::active", self._on_client_update)

            self._on_location_update(self.simple)

    def get_data(self) -> None:
        Geoclue.Simple.new(
            app.get_default_app_id(),
            Geoclue.AccuracyLevel.EXACT,
            None,
            self._on_simple_ready,
        )

    def start(self) -> None:
        self._running = True
        self.location_info = {}
        self.get_data()

    def _send_location(self) -> None:
        if self.location_info == self._data:
            return
        if "timestamp" in self.location_info and "timestamp" in self._data:
            last_data = self.location_info.copy()
            del last_data["timestamp"]
            new_data = self._data.copy()
            del new_data["timestamp"]
            if last_data == new_data:
                return
        self.location_info = self._data.copy()
        info = LocationData(**self._data)
        self._emit(info)

    @staticmethod
    def _timestamp_to_string(timestamp: float) -> str:
        utc_datetime = datetime.fromtimestamp(timestamp, UTC)
        return utc_datetime.strftime("%Y-%m-%dT%H:%MZ")



# Copyright (C) 2003-2014 Yann Leboulanger <asterix AT lagaule.org>
# Copyright (C) 2005-2006 Nikos Kouremenos <kourem AT gmail.com>
# Copyright (C) 2007 Jean-Marie Traissard <jim AT lapin.org>
# Copyright (C) 2008 Mateusz Biliński <mateusz AT bilinski.it>
# Copyright (C) 2008 Thorsten P. 'dGhvcnN0ZW5wIEFUIHltYWlsIGNvbQ==\n'.decode("base64")  # noqa: E501
# Copyright (C) 2018 Philipp Hörist <philipp AT hoerist.com>
#
# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import cast

import ctypes
import ctypes.util
import logging
import sys
import time

from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject

from gajim.common import app
from gajim.common.const import Display
from gajim.common.const import IdleState

log = logging.getLogger('gajim.c.idle')


class IdleMonitor:
    def __init__(self):
        self._extended_away = False

    def get_idle_sec(self) -> int:
        raise NotImplementedError

    def set_extended_away(self, state: bool) -> None:
        self._extended_away = state

    def is_extended_away(self) -> bool:
        return self._extended_away


class DBusFreedesktop(IdleMonitor):

    def __init__(self) -> None:
        IdleMonitor.__init__(self)
        self._last_idle_time = 0

        log.debug('Connecting to org.freedesktop.ScreenSaver')
        self._dbus_proxy = Gio.DBusProxy.new_for_bus_sync(
            Gio.BusType.SESSION,
            Gio.DBusProxyFlags.NONE,
            None,
            'org.freedesktop.ScreenSaver',
            '/org/freedesktop/ScreenSaver',
            'org.freedesktop.ScreenSaver',
            None
        )
        log.debug('Connected')

        # Only the following call will trigger exceptions if the D-Bus
        # interface/method/... does not exist. Using the failing method
        # for class init to allow other idle monitors to be used on failure.
        self._get_idle_sec_fail()
        log.debug('Test successful')

    def _get_idle_sec_fail(self) -> int:
        (idle_time,) = cast(tuple[int], self._dbus_proxy.call_sync(
            'GetSessionIdleTime',
            None,
            Gio.DBusCallFlags.NO_AUTO_START,
            -1,
            None))

        return idle_time // 1000

    def get_idle_sec(self) -> int:
        try:
            self._last_idle_time = self._get_idle_sec_fail()
        except GLib.Error as error:
            log.warning(
                'org.freedesktop.ScreenSaver.GetSessionIdleTime() failed: %s',
                error)

        return self._last_idle_time


class DBusGnome(IdleMonitor):

    def __init__(self) -> None:
        IdleMonitor.__init__(self)
        self._last_idle_time = 0

        log.debug('Connecting to org.gnome.Mutter.IdleMonitor')
        self._dbus_proxy = Gio.DBusProxy.new_for_bus_sync(
            Gio.BusType.SESSION,
            Gio.DBusProxyFlags.NONE,
            None,
            'org.gnome.Mutter.IdleMonitor',
            '/org/gnome/Mutter/IdleMonitor/Core',
            'org.gnome.Mutter.IdleMonitor',
            None
        )
        log.debug('Connected')

        # Only the following call will trigger exceptions if the D-Bus
        # interface/method/... does not exist. Using the failing method
        # for class init to allow other idle monitors to be used on failure.
        self._get_idle_sec_fail()
        log.debug('Test successful')

    def _get_idle_sec_fail(self) -> int:
        (idle_time,) = cast(tuple[int], self._dbus_proxy.call_sync(
            'GetIdletime',
            None,
            Gio.DBusCallFlags.NO_AUTO_START,
            -1,
            None))

        return idle_time // 1000

    def get_idle_sec(self) -> int:
        try:
            self._last_idle_time = self._get_idle_sec_fail()
        except GLib.Error as error:
            log.warning(
                'org.gnome.Mutter.IdleMonitor.GetIdletime() failed: %s',
                error)

        return self._last_idle_time


class Xss(IdleMonitor):
    def __init__(self) -> None:
        IdleMonitor.__init__(self)

        class XScreenSaverInfo(ctypes.Structure):
            _fields_ = [
                ('window', ctypes.c_ulong),
                ('state', ctypes.c_int),
                ('kind', ctypes.c_int),
                ('til_or_since', ctypes.c_ulong),
                ('idle', ctypes.c_ulong),
                ('eventMask', ctypes.c_ulong)
            ]

        XScreenSaverInfo_p = ctypes.POINTER(XScreenSaverInfo)

        display_p = ctypes.c_void_p
        xid = ctypes.c_ulong
        c_int_p = ctypes.POINTER(ctypes.c_int)

        lib_x11_path = ctypes.util.find_library('X11')
        if lib_x11_path is None:
            raise OSError('libX11 could not be found.')

        lib_x11 = ctypes.cdll.LoadLibrary(lib_x11_path)
        lib_x11.XOpenDisplay.restype = display_p
        lib_x11.XOpenDisplay.argtypes = (ctypes.c_char_p,)
        lib_x11.XDefaultRootWindow.restype = xid
        lib_x11.XDefaultRootWindow.argtypes = (display_p,)

        lib_xss_path = ctypes.util.find_library('Xss')
        if lib_xss_path is None:
            raise OSError('libXss could not be found.')

        self._lib_xss = ctypes.cdll.LoadLibrary(lib_xss_path)
        self._lib_xss.XScreenSaverQueryExtension.argtypes = (
            display_p, c_int_p, c_int_p)
        self._lib_xss.XScreenSaverAllocInfo.restype = XScreenSaverInfo_p
        self._lib_xss.XScreenSaverQueryInfo.argtypes = (
            display_p, xid, XScreenSaverInfo_p)

        self._dpy_p = lib_x11.XOpenDisplay(None)
        if self._dpy_p is None:
            raise OSError('Could not open X Display.')

        _event_basep = ctypes.c_int()
        _error_basep = ctypes.c_int()
        extension = self._lib_xss.XScreenSaverQueryExtension(
            self._dpy_p, ctypes.byref(_event_basep), ctypes.byref(_error_basep))
        if extension == 0:
            raise OSError('XScreenSaver Extension not available on display.')

        self._xss_info_p = self._lib_xss.XScreenSaverAllocInfo()
        if self._xss_info_p is None:
            raise OSError('XScreenSaverAllocInfo: Out of Memory.')

        self.root_window = lib_x11.XDefaultRootWindow(self._dpy_p)

    def get_idle_sec(self) -> int:
        info = self._lib_xss.XScreenSaverQueryInfo(
            self._dpy_p, self.root_window, self._xss_info_p)
        if info == 0:
            return info
        return self._xss_info_p.contents.idle // 1000


class Windows(IdleMonitor):
    def __init__(self) -> None:
        IdleMonitor.__init__(self)
        self._OpenInputDesktop = ctypes.windll.user32.OpenInputDesktop
        self._CloseDesktop = ctypes.windll.user32.CloseDesktop
        self._SystemParametersInfo = ctypes.windll.user32.SystemParametersInfoW
        self._GetTickCount = ctypes.windll.kernel32.GetTickCount
        self._GetLastInputInfo = ctypes.windll.user32.GetLastInputInfo

        self._locked_time = None

        class LastInputInfo(ctypes.Structure):
            _fields_ = [
                ('cbSize', ctypes.c_uint),
                ('dwTime', ctypes.c_uint)
            ]

        self._lastInputInfo = LastInputInfo()
        self._lastInputInfo.cbSize = ctypes.sizeof(self._lastInputInfo)

    def get_idle_sec(self) -> int:
        self._GetLastInputInfo(ctypes.byref(self._lastInputInfo))
        return int(self._GetTickCount() - self._lastInputInfo.dwTime) // 1000

    def set_extended_away(self, state: bool) -> None:
        raise NotImplementedError

    def is_extended_away(self) -> bool:
        # Check if Screen Saver is running
        # 0x72 is SPI_GETSCREENSAVERRUNNING
        saver_runing = ctypes.c_int(0)
        info = self._SystemParametersInfo(
            0x72, 0, ctypes.byref(saver_runing), 0)
        if info and saver_runing.value:
            return True

        # Check if Screen is locked
        # Also a UAC prompt counts as locked
        # So just return True if we are more than 10 seconds locked
        desk = self._OpenInputDesktop(0, False, 0)
        unlocked = bool(desk)
        self._CloseDesktop(desk)

        if unlocked:
            self._locked_time = None
            return False

        if self._locked_time is None:
            self._locked_time = time.time()
            return False

        threshold = time.time() - 10
        return threshold > self._locked_time


class IdleMonitorManager(GObject.Object):

    __gsignals__ = {
        'state-changed': (
            GObject.SignalFlags.RUN_LAST | GObject.SignalFlags.ACTION,
            None,
            ()
        )}

    def __init__(self):
        GObject.Object.__init__(self)
        self.set_interval()
        self._state = IdleState.AWAKE
        self._idle_monitor = self._get_idle_monitor()

        if self.is_available():
            GLib.timeout_add_seconds(5, self._poll)

    def set_interval(self,
                     away_interval: int = 60,
                     xa_interval: int = 120) -> None:

        log.info('Set interval: away: %s, xa: %s',
                 away_interval, xa_interval)
        self._away_interval = away_interval
        self._xa_interval = xa_interval

    def set_extended_away(self, state: bool) -> None:
        if self._idle_monitor is None:
            raise ValueError('No idle monitor available')

        self._idle_monitor.set_extended_away(state)

    def is_available(self) -> bool:
        return self._idle_monitor is not None

    @property
    def state(self) -> IdleState:
        if not self.is_available():
            return IdleState.UNKNOWN
        return self._state

    def is_xa(self) -> bool:
        return self.state == IdleState.XA

    def is_away(self) -> bool:
        return self.state == IdleState.AWAY

    def is_awake(self) -> bool:
        return self.state == IdleState.AWAKE

    def is_unknown(self) -> bool:
        return self.state == IdleState.UNKNOWN

    @staticmethod
    def _get_idle_monitor() -> IdleMonitor | None:
        if sys.platform == 'win32':
            return Windows()

        try:
            return DBusFreedesktop()
        except GLib.Error as error:
            log.info('Idle time via org.freedesktop.Screensaver '
                     'not available: %s', error)

        try:
            return DBusGnome()
        except GLib.Error as error:
            log.info('Idle time via org.gnome.Mutter.IdleMonitor '
                     'not available: %s', error)

        if app.is_display(Display.WAYLAND):
            return None

        try:
            return Xss()
        except OSError as error:
            log.info('Idle time via XScreenSaverInfo not available: %s', error)

        return None

    def get_idle_sec(self) -> int:
        if self._idle_monitor is None:
            raise ValueError('No idle monitor available')
        return self._idle_monitor.get_idle_sec()

    def _poll(self) -> bool:
        '''
        Check to see if we should change state
        '''
        assert self._idle_monitor is not None

        if self._idle_monitor.is_extended_away():
            log.info('Extended Away: Screensaver or Locked Screen')
            self._set_state(IdleState.XA)
            return True

        idle_time = self.get_idle_sec()

        # xa is stronger than away so check for xa first
        if idle_time > self._xa_interval:
            self._set_state(IdleState.XA)
        elif idle_time > self._away_interval:
            self._set_state(IdleState.AWAY)
        else:
            self._set_state(IdleState.AWAKE)
        return True

    def _set_state(self, state: IdleState) -> None:
        if self._state == state:
            return

        self._state = state
        log.info('State changed: %s', state)
        self.emit('state-changed')


Monitor = IdleMonitorManager()



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from typing import Any

from pathlib import Path

from gajim.common import configpaths

from gajim.gtk.builder import GajimBuilder

from .plugins_i18n import _
from .plugins_i18n import DOMAIN


class GajimPluginActivateException(Exception):
    """
    Raised when activation failed
    """


def get_builder(
    file_name: str, instance: Any = None, widgets: list[str] | None = None
) -> GajimBuilder:
    return GajimBuilder(file_name, instance, widgets, domain=DOMAIN, gettext_=_)


def is_shipped_plugin(path: Path) -> bool:
    base = configpaths.get("PLUGINS_BASE")
    if not base.exists():
        return False
    plugin_parent = path.parent
    return base.samefile(plugin_parent)



#
# SPDX-License-Identifier: GPL-3.0-only

import logging
from pathlib import Path

from gi.repository import Adw
from gi.repository import Gtk
from nbxmpp.const import Affiliation
from nbxmpp.errors import StanzaError
from nbxmpp.modules.vcard_temp import VCard
from nbxmpp.namespaces import Namespace
from nbxmpp.structs import MucSubject
from nbxmpp.task import Task

from gajim.common import app
from gajim.common.const import AvatarSize
from gajim.common.i18n import _
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.util.jid import validate_jid

from gajim.gtk.alert import InformationAlertDialog
from gajim.gtk.avatar_selector import AvatarSelector
from gajim.gtk.builder import get_builder
from gajim.gtk.filechoosers import AvatarFileChooserButton
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.window import get_app_window

log = logging.getLogger("gajim.gtk.groupchat_manage")


class GroupchatManage(Gtk.Box, SignalManager):
    def __init__(
        self,
        account: str,
        contact: GroupchatContact,
    ) -> None:
        Gtk.Box.__init__(self)
        SignalManager.__init__(self)

        self._account = account
        self._client = app.get_client(account)
        self._contact = contact
        self._contact.connect("room-subject", self._on_room_subject)

        self._room_config_form = None

        self._ui = get_builder("groupchat_manage.ui")
        self.append(self._ui.scrolled)

        self._avatar_chooser_button = AvatarFileChooserButton(
            tooltip=_("Change your profile picture"),
            icon_name="lucide-square-pen-symbolic",
        )
        self._avatar_chooser_button.set_halign(Gtk.Align.END)
        self._avatar_chooser_button.set_valign(Gtk.Align.END)
        self._avatar_chooser_button.set_visible(False)
        self._avatar_chooser_button.add_css_class("opaque")

        self._ui.avatar_overlay.add_overlay(self._avatar_chooser_button)

        self._connect(self._ui.remove_avatar_button, "clicked", self._on_remove_avatar)
        self._connect(
            self._ui.muc_description_entry_row, "changed", self._on_name_desc_changed
        )
        self._connect(
            self._ui.muc_name_entry_row, "changed", self._on_name_desc_changed
        )
        self._connect(self._ui.destroy_muc_button, "clicked", self._on_destroy_clicked)
        self._connect(
            self._ui.manage_save_button, "clicked", self._on_manage_save_clicked
        )
        self._connect(
            self._ui.subject_change_button, "clicked", self._on_subject_change_clicked
        )
        self._connect(
            self._ui.avatar_update_button, "clicked", self._on_avatar_update_clicked
        )
        self._connect(
            self._ui.avatar_cancel_button, "clicked", self._on_avatar_cancel_clicked
        )
        self._connect(
            self._ui.destroy_alternate_entry,
            "changed",
            self._on_destroy_alternate_changed,
        )
        self._connect(
            self._ui.destroy_cancel_button, "clicked", self._on_destroy_cancelled
        )
        self._connect(self._ui.destroy_button, "clicked", self._on_destroy_confirmed)
        self._connect(
            self._ui.subject_textview.get_buffer(),
            "changed",
            self._on_subject_text_changed,
        )
        self._connect(
            self._avatar_chooser_button, "path-picked", self._on_avatar_picked
        )

        self._avatar_selector = AvatarSelector()
        self._avatar_selector.set_size_request(500, 500)
        self._ui.avatar_selector_grid.attach(self._avatar_selector, 0, 1, 1, 1)

        self._prepare_subject()
        self._prepare_manage()

    def do_unroot(self) -> None:
        self._disconnect_all()
        Gtk.Box.do_unroot(self)
        del self._avatar_selector
        del self._avatar_chooser_button
        del self._room_config_form
        app.check_finalize(self)

    @property
    def disco_info(self):
        return app.storage.cache.get_last_disco_info(self._contact.jid)

    def _prepare_subject(self) -> None:
        text = ""
        if self._contact.subject is not None:
            text = self._contact.subject.text

        self._ui.subject_textview.get_buffer().set_text(text)
        self._ui.subject_textview.set_sensitive(self._is_subject_change_allowed())

    def _is_subject_change_allowed(self) -> bool:
        if not self._contact.is_joined:
            return False

        self_contact = self._contact.get_self()
        if self_contact is None:
            return False

        if self_contact.affiliation in (Affiliation.OWNER, Affiliation.ADMIN):
            return True

        if self.disco_info is None:
            return False
        return self.disco_info.muc_subjectmod or False

    def _on_subject_text_changed(self, buffer_: Gtk.TextBuffer) -> None:
        text = buffer_.get_text(buffer_.get_start_iter(), buffer_.get_end_iter(), False)

        assert self._contact.subject is not None

        self._ui.subject_change_button.set_sensitive(text != self._contact.subject.text)

    def _on_subject_change_clicked(self, _button: Gtk.Button) -> None:
        buffer_ = self._ui.subject_textview.get_buffer()
        subject = buffer_.get_text(
            buffer_.get_start_iter(), buffer_.get_end_iter(), False
        )
        self._client.get_module("MUC").set_subject(self._contact.jid, subject)

    def _on_room_subject(
        self, _contact: GroupchatContact, _signal_name: str, subject: MucSubject
    ) -> None:
        self._ui.subject_textview.get_buffer().set_text(subject.text)

        assert self._contact.subject is not None
        self._ui.subject_change_button.set_sensitive(
            subject.text != self._contact.subject.text
        )

    def _prepare_manage(self) -> None:
        joined = self._contact.is_joined
        vcard_support = False

        if self.disco_info is not None:
            vcard_support = self.disco_info.supports(Namespace.VCARD)

            self._ui.muc_name_entry_row.set_text(self.disco_info.muc_name or "")
            self._ui.muc_description_entry_row.set_text(
                self.disco_info.muc_description or ""
            )

        self.update_avatar()

        self_contact = self._contact.get_self()
        if not joined or self_contact is None:
            return

        if vcard_support and self_contact.affiliation.is_owner:
            self._avatar_chooser_button.set_visible(True)
            self._ui.remove_avatar_button.set_visible(True)

        if self_contact.affiliation.is_owner:
            self._client.get_module("MUC").request_config(
                self._contact.jid, callback=self._on_manage_form_received
            )
            self._ui.destroy_muc_button.set_sensitive(True)

    def _on_manage_form_received(self, task: Task) -> None:
        try:
            result = task.finish()
        except StanzaError as error:
            log.info(error)
            return

        self._ui.muc_name_entry_row.set_sensitive(True)
        self._ui.muc_description_entry_row.set_sensitive(True)
        self._room_config_form = result.form

    def _on_name_desc_changed(self, _entry: Adw.EntryRow) -> None:
        if self.disco_info is not None:
            disco_name = self.disco_info.muc_name or ""
            disco_desc = self.disco_info.muc_description or ""
            name = self._ui.muc_name_entry_row.get_text()
            desc = self._ui.muc_description_entry_row.get_text()

            self._ui.manage_save_button.set_sensitive(
                desc != disco_desc or name != disco_name
            )

    def _on_manage_save_clicked(self, _button: Gtk.Button) -> None:
        if self._room_config_form is not None:
            name = self._ui.muc_name_entry_row.get_text()
            desc = self._ui.muc_description_entry_row.get_text()
            try:
                name_field = self._room_config_form["muc#roomconfig_roomname"]
                desc_field = self._room_config_form["muc#roomconfig_roomdesc"]
            except KeyError:
                pass
            else:
                name_field.value = name
                desc_field.value = desc

            self._client.get_module("MUC").set_config(
                self._contact.jid, self._room_config_form
            )

    def _on_avatar_cancel_clicked(self, _button: Gtk.Button) -> None:
        self._ui.stack.set_visible_child_name("manage")

    def _on_remove_avatar(self, _button: Gtk.Button) -> None:
        vcard = VCard()
        self._client.get_module("VCardTemp").set_vcard(
            vcard, jid=self._contact.jid, callback=self._on_upload_avatar_result
        )

    def _on_avatar_picked(
        self, _button: AvatarFileChooserButton, paths: list[Path]
    ) -> None:
        self._avatar_selector.prepare_crop_area(str(paths[0]))
        self._ui.avatar_update_button.set_sensitive(
            self._avatar_selector.get_prepared()
        )
        self._ui.stack.set_visible_child_name("avatar")

    def _on_upload_avatar_result(self, task: Task) -> None:
        try:
            task.finish()
        except Exception as error:
            InformationAlertDialog(
                _("Uploading Avatar Failed"),
                _("Uploading avatar image failed: %s") % error,
            )

    def _on_avatar_update_clicked(self, _button: Gtk.Button) -> None:
        success, data, _w, _h = self._avatar_selector.get_avatar_bytes()
        if not success:
            InformationAlertDialog(
                _("Loading Avatar Failed"), _("Loading avatar image failed")
            )
            return

        assert data
        sha = app.app.avatar_storage.save_avatar(data)
        if sha is None:
            InformationAlertDialog(
                _("Saving Avatar Failed"), _("Saving avatar image failed")
            )
            return

        vcard = VCard()
        vcard.set_avatar(data, "image/png")

        self._client.get_module("VCardTemp").set_vcard(
            vcard, jid=self._contact.jid, callback=self._on_upload_avatar_result
        )
        self._ui.stack.set_visible_child_name("manage")

    def update_avatar(self):
        texture = app.app.avatar_storage.get_muc_texture(
            self._account,
            self._contact.jid,
            AvatarSize.GROUP_INFO,
            self.get_scale_factor(),
        )

        self._ui.avatar_button_image.set_pixel_size(AvatarSize.GROUP_INFO)
        self._ui.avatar_button_image.set_from_paintable(texture)

    def _on_destroy_clicked(self, _button: Gtk.Button) -> None:
        self._ui.stack.set_visible_child_name("destroy")
        self._ui.destroy_reason_entry.grab_focus()

    def _on_destroy_alternate_changed(self, entry: Gtk.Entry) -> None:
        jid = entry.get_text()
        if jid:
            try:
                jid = validate_jid(jid)
            except Exception:
                icon = "lucide-circle-alert-symbolic"
                text = _("Invalid XMPP Address")
                self._ui.destroy_alternate_entry.set_icon_from_icon_name(
                    Gtk.EntryIconPosition.SECONDARY, icon
                )
                self._ui.destroy_alternate_entry.set_icon_tooltip_text(
                    Gtk.EntryIconPosition.SECONDARY, text
                )
                self._ui.destroy_button.set_sensitive(False)
                return
        self._ui.destroy_alternate_entry.set_icon_from_icon_name(
            Gtk.EntryIconPosition.SECONDARY, None
        )
        self._ui.destroy_button.set_sensitive(True)

    def _on_destroy_confirmed(self, _button: Gtk.Button) -> None:
        reason = self._ui.destroy_reason_entry.get_text()
        alternate_jid = self._ui.destroy_alternate_entry.get_text()
        self._client.get_module("MUC").destroy(self._contact.jid, reason, alternate_jid)

        window = get_app_window("GroupchatDetails")
        assert window is not None
        window.close()

    def _on_destroy_cancelled(self, _button: Gtk.Button) -> None:
        self._ui.stack.set_visible_child_name("manage")



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any

import logging
import os
import pickle
import types
from collections.abc import Generator

from gajim.common import configpaths
from gajim.common.events import ApplicationEvent
from gajim.common.ged import HandlerFuncT
from gajim.common.types import PluginExtensionPoints

from .manifest import PluginManifest

log = logging.getLogger("gajim.p.plugin")


class GajimPlugin:  # noqa: PLW1641
    """
    Base class for implementing Gajim plugins.
    """

    __path__: str = ""

    encryption_name: str = ""
    manifest: PluginManifest

    gui_extension_points: PluginExtensionPoints = {}
    """
    Extension points that plugin wants to connect with and handlers to be used.

    Keys of this string should be strings with name of GUI extension point
    to handles. Values should be 2-element tuples with references to handling
    functions. First function will be used to connect plugin with extpoint,
    the second one to successfully disconnect from it. Connecting takes places
    when plugin is activated and extpoint already exists, or when plugin is
    already activated but extpoint is being created (eg. chat window opens).
    Disconnecting takes place when plugin is deactivated and extpoint exists
    or when extpoint is destroyed and plugin is activate (eg. chat window
    closed).
    """
    config_default_values: dict[str, tuple[Any, str]] = {}
    """
    Default values for keys that should be stored in plug-in config.

    This dict is used when when someone calls for config option but it has not
    been set yet.

    Values are tuples: (default_value, option_description). The first one can
    be anything (this is the advantage of using shelve/pickle instead of
    custom-made     config I/O handling); the second one should be str (gettext
    can be used if need and/or translation is planned).
    """
    events_handlers: dict[str, tuple[int, HandlerFuncT]] = {}
    """
    Dictionary with events handlers.

    Keys are event names. Values should be 2-element tuples with handler
    priority as first element and reference to handler function as second
    element. Priority is integer. See `ged` module for predefined priorities
    like `ged.PRECORE`, `ged.CORE` or `ged.POSTCORE`.
    """
    events: list[ApplicationEvent] = []
    """
    New network event classes to be registered in Network Events Controller.
    """
    modules: list[types.ModuleType] = []

    def __init__(self) -> None:
        self.config = GajimPluginConfig(self)
        self.activatable = True
        self.active = False
        self.available_text = ""
        self.load_config()
        self.config_dialog = None
        self.init()

    def save_config(self) -> None:
        self.config.save()

    def load_config(self) -> None:
        self.config.load()

    def __eq__(self, plugin: Any) -> bool:
        if not isinstance(plugin, GajimPlugin):
            return False
        return self.manifest.short_name == plugin.manifest.short_name

    def __ne__(self, plugin: Any) -> bool:
        return self.manifest.short_name != plugin.manifest.short_name

    def local_file_path(self, file_name: str) -> str:
        return os.path.join(self.__path__, file_name)

    def init(self) -> None:
        pass

    def activate(self) -> None:
        pass

    def deactivate(self) -> None:
        pass

    def activate_encryption(self, chat_control: Any) -> bool:
        raise NotImplementedError


class GajimPluginConfig:
    def __init__(self, plugin: GajimPlugin) -> None:
        self.plugin = plugin
        self.FILE_PATH = (
            configpaths.get("PLUGINS_CONFIG_DIR") / self.plugin.manifest.short_name
        )
        self.data: dict[str, Any] = {}

    def __getitem__(self, key: str) -> None:
        if key not in self.data:
            self.data[key] = self.plugin.config_default_values[key][0]
            self.save()

        return self.data[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.data[key] = value
        self.save()

    def __delitem__(self, key: str) -> None:
        del self.data[key]
        self.save()

    def __contains__(self, key: str) -> bool:
        return key in self.data

    def __iter__(self) -> Generator[str, Any, None]:
        yield from self.data

    def keys(self):
        return self.data.keys()

    def items(self):
        return self.data.items()

    def save(self) -> None:
        with open(self.FILE_PATH, "wb") as fd:
            pickle.dump(self.data, fd)

    def load(self) -> None:
        if not self.FILE_PATH.is_file():
            self.data = {}
            self.save()
            return
        with open(self.FILE_PATH, "rb") as fd:
            try:
                self.data = pickle.load(fd)
            except Exception:
                try:
                    import shelve

                    shelf = shelve.open(str(self.FILE_PATH))
                    for key, value in shelf.items():
                        self.data[key] = value
                    if not isinstance(self.data, dict):  # pyright: ignore
                        raise GajimPluginException
                    shelf.close()
                    self.save()
                except Exception:
                    filepath_bak = self.FILE_PATH.with_suffix("bak")
                    log.warning(
                        "%s plugin config file not readable. Saving it as "
                        "%s and creating a new one",
                        self.plugin.manifest.short_name,
                        filepath_bak,
                    )
                    if filepath_bak.exists():
                        filepath_bak.unlink()

                    self.FILE_PATH.rename(f"{self.FILE_PATH}.bak")
                    self.data = {}
                    self.save()


class GajimPluginException(Exception):
    pass


class GajimPluginInitError(GajimPluginException):
    pass



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import cast
from typing import Literal

import logging
from collections import defaultdict
from collections.abc import Generator
from datetime import datetime

from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Graphene
from gi.repository import Gtk
from nbxmpp.errors import StanzaError
from nbxmpp.protocol import JID
from nbxmpp.structs import MucSubject

import gajim.common.storage.archive.models as mod
from gajim.common import app
from gajim.common import events
from gajim.common import ged
from gajim.common import types
from gajim.common.const import Direction
from gajim.common.helpers import to_user_string
from gajim.common.modules.chat_markers import DisplayedMarkerData
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.modules.contacts import GroupchatParticipant
from gajim.common.modules.httpupload import HTTPFileTransfer
from gajim.common.storage.archive.const import ChatDirection
from gajim.common.storage.archive.const import MessageType
from gajim.common.storage.archive.models import Message
from gajim.common.types import ChatContactT
from gajim.common.util.datetime import get_start_of_day

from gajim.gtk.conversation.rows.base import BaseRow
from gajim.gtk.conversation.rows.call import CallRow
from gajim.gtk.conversation.rows.command_output import CommandOutputRow
from gajim.gtk.conversation.rows.date import DateRow
from gajim.gtk.conversation.rows.displayed import DisplayedRow
from gajim.gtk.conversation.rows.encryption_info import EncryptionInfoRow
from gajim.gtk.conversation.rows.file_transfer import FileTransferRow
from gajim.gtk.conversation.rows.file_transfer_jingle import FileTransferJingleRow
from gajim.gtk.conversation.rows.info import InfoMessage
from gajim.gtk.conversation.rows.message import MessageRow
from gajim.gtk.conversation.rows.muc_join_left import MUCJoinLeft
from gajim.gtk.conversation.rows.muc_subject import MUCSubject
from gajim.gtk.conversation.rows.scroll_hint import ScrollHintRow
from gajim.gtk.conversation.rows.user_status import UserStatus
from gajim.gtk.conversation.rows.widgets import MessageRowActions
from gajim.gtk.util.misc import check_finalize
from gajim.gtk.util.misc import iterate_listbox_children

log = logging.getLogger("gajim.gtk.conversation_view")


class ConversationView(Gtk.ScrolledWindow):
    __gsignals__ = {
        "request-history": (
            GObject.SignalFlags.RUN_LAST | GObject.SignalFlags.ACTION,
            None,
            (bool,),
        ),
        "autoscroll-changed": (GObject.SignalFlags.RUN_LAST, None, (bool,)),
    }

    def __init__(self, message_row_actions: MessageRowActions) -> None:
        Gtk.ScrolledWindow.__init__(self)

        self.set_overlay_scrolling(False)
        self.add_css_class("scrolled-no-border")
        self.add_css_class("no-scroll-indicator")
        self.add_css_class("scrollbar-style")
        self.set_vexpand(True)

        self.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)

        # This is a workaround: as soon as a line break occurs in Gtk.TextView
        # with word-char wrapping enabled, a hyphen character is automatically
        # inserted before the line break. This triggers the hscrollbar to show,
        # see: https://gitlab.gnome.org/GNOME/gtk/-/issues/2384
        # Using set_hscroll_policy(Gtk.Scrollable.Policy.NEVER) would cause bad
        # performance during resize, and prevent the window from being shrunk
        # horizontally under certain conditions (applies to GroupchatControl)
        self.get_hscrollbar().set_visible(False)

        self._message_row_actions = message_row_actions

        self._contact: ChatContactT | None = None
        self._client = None

        self._list_box = Gtk.ListBox()

        # Keeps track of the number of rows shown in ConversationView
        self._row_count: int = 0
        self._max_row_count: int = 100

        # Keeps track of date rows we have added to the list
        self._active_date_rows: set[datetime] = set()

        self._message_id_row_map: dict[str, MessageRow] = {}
        self._stanza_id_row_map: dict[str, MessageRow] = {}

        self._scroll_hint_row = None

        self._current_upper: float = 0
        self._autoscroll: bool = True
        self._request_history_at_upper: float | None = None
        self._upper_complete: bool = False
        self._lower_complete: bool = True
        self._requesting: str | None = None
        self._block_signals = False

        self._signal_handlers_enabled = False
        self._signal_handler_ids = (0, 0)
        # Maps occupnat id -> markers
        self._dm_occupant_markers: dict[str | None, DisplayedMarkerData] = {}
        # Maps message ids -> occupant id -> markers
        self._dm_id_occupant_markers: dict[
            str, dict[str | None, DisplayedMarkerData]
        ] = defaultdict(dict)
        # Maps message ids -> DisplayedRows
        self._dm_rows: dict[str, DisplayedRow] = {}

        self._last_occupant_messages: dict[str, mod.Message] = {}

        self.set_child(self._list_box)

        app.ged.register_event_handler(
            "register-actions", ged.GUI1, self._on_register_actions
        )

    def _on_register_actions(self, _event: events.RegisterActions) -> None:
        app.window.get_action("scroll-view-up").connect(
            "activate", self._on_scroll_view
        )
        app.window.get_action("scroll-view-down").connect(
            "activate", self._on_scroll_view
        )

    def copy_selected_messages(self) -> None:
        format_string = app.settings.get("date_time_format")
        selection_text = ""
        for row in cast(list[BaseRow], self._list_box.get_selected_rows()):
            if isinstance(row, MessageRow):
                timestamp_formatted = row.timestamp.strftime(format_string)
                selection_text += (
                    f"{timestamp_formatted} - {row.name}:\n{row.get_text()}\n"
                )

        self.get_clipboard().set(selection_text)

        self.disable_row_selection()

    def enable_row_selection(self, pk: int | None) -> None:
        self._list_box.set_selection_mode(Gtk.SelectionMode.MULTIPLE)

        if pk is not None:
            row = self.get_row_by_pk(pk)
            self._list_box.select_row(row)

        for row in self.iter_rows():
            row.enable_selection_mode()

    def disable_row_selection(self) -> None:
        self._list_box.unselect_all()
        self._list_box.set_selection_mode(Gtk.SelectionMode.NONE)

        for row in self.iter_rows():
            row.disable_selection_mode()

    def _on_scroll_view(self, action: Gio.SimpleAction, _param: Literal[None]) -> None:
        action_name = action.get_name()
        if action_name == "scroll-view-down":
            self.emit("scroll-child", Gtk.ScrollType.PAGE_DOWN, False)

        elif action_name == "scroll-view-up":
            self.emit("scroll-child", Gtk.ScrollType.PAGE_UP, False)

    def _enable_signal_handlers(self, enable: bool) -> None:
        if self._signal_handlers_enabled == enable:
            return

        vadjustment = self.get_vadjustment()

        if enable:
            upper_id = vadjustment.connect("notify::upper", self._on_adj_upper_changed)
            value_id = vadjustment.connect("notify::value", self._on_adj_value_changed)
            self._signal_handler_ids = (upper_id, value_id)
        else:
            upper_id, value_id = self._signal_handler_ids
            vadjustment.disconnect(upper_id)
            vadjustment.disconnect(value_id)

        self._signal_handlers_enabled = enable

    def clear(self) -> None:
        app.settings.disconnect_signals(self)
        self._enable_signal_handlers(False)
        self._reset()

        self._contact = None
        self._client = None

    def reload(self) -> None:
        if self._contact is None:
            return
        self.switch_contact(self._contact)

    def switch_contact(self, contact: ChatContactT) -> None:
        self._contact = contact
        self._client = app.get_client(contact.account)

        self._enable_signal_handlers(False)
        self._block_signals = True
        self._reset()

        self.disable_row_selection()

        if isinstance(self._contact, GroupchatParticipant):
            show_markers = self._contact.room.settings.get("send_marker")
        else:
            show_markers = self._contact.settings.get("send_marker")

        if show_markers:
            self._load_displayed_marker()

        self._scroll_hint_row = ScrollHintRow(self._contact.account)
        self._list_box.append(self._scroll_hint_row)

        app.settings.disconnect_signals(self)

        app.settings.connect_signal(
            "print_join_left",
            self._on_contact_setting_changed,
            account=contact.account,
            jid=contact.jid,
        )

        app.settings.connect_signal(
            "print_status",
            self._on_contact_setting_changed,
            account=contact.account,
            jid=contact.jid,
        )

        self._block_signals = False
        self._enable_signal_handlers(True)

    def get_autoscroll(self) -> bool:
        return self._autoscroll

    def block_signals(self, value: bool) -> None:
        self._block_signals = value

    def _emit(self, signal_name: str, *args: Any) -> None:
        if not self._block_signals:
            log.debug("emit %s, %s", signal_name, args)
            GLib.idle_add(self.emit, signal_name, *args)

    def _reset_list_box(self) -> None:
        self._list_box.set_selection_mode(Gtk.SelectionMode.NONE)
        # Performance: Disable sort function before removing all rows
        self._list_box.set_sort_func(None)
        check_finalize(self._list_box, only_children=True)
        self._list_box.remove_all()

        self._list_box.set_sort_func(self._sort_func)

    def _reset(self) -> None:
        self._current_upper = 0
        self._autoscroll = True
        self._request_history_at_upper = None
        self._upper_complete = False
        self._lower_complete = True
        self._requesting = None
        self.set_history_complete(True, False)

        self._reset_list_box()

        self._row_count = 0
        self._active_date_rows = set()
        self._message_id_row_map = {}
        self._stanza_id_row_map = {}
        self._scroll_hint_row = None

        self._dm_id_occupant_markers.clear()
        self._dm_occupant_markers.clear()
        self._dm_rows.clear()

    def reset(self) -> None:
        assert self._contact is not None
        self.switch_contact(self._contact)

    def set_history_complete(self, before: bool, complete: bool) -> None:
        if before:
            self._upper_complete = complete
            if self._scroll_hint_row is None:
                return
            self._scroll_hint_row.set_history_complete(complete)
        else:
            self._lower_complete = complete

    def get_lower_complete(self) -> bool:
        return self._lower_complete

    def _scroll_to_pos_idle(self, adj: Gtk.Adjustment, value: float) -> None:
        if value == -1:
            # scroll to end
            value = adj.get_upper() - adj.get_page_size()

        GLib.idle_add(adj.set_value, value)

    def _on_adj_upper_changed(
        self, adj: Gtk.Adjustment, _pspec: GObject.ParamSpec
    ) -> None:
        upper = adj.get_upper()
        diff = upper - self._current_upper

        if diff != 0:
            self._current_upper = upper
            if self._autoscroll:
                self._scroll_to_pos_idle(adj, -1)
            else:
                # Workaround
                # https://gitlab.gnome.org/GNOME/gtk/merge_requests/395
                self.set_kinetic_scrolling(True)
                if self._requesting == "before":
                    self._scroll_to_pos_idle(adj, adj.get_value() + diff)

        if upper == adj.get_page_size():
            # There is no scrollbar
            if not self._block_signals:
                self._emit("request-history", True)
            self._lower_complete = True
            self._autoscroll = True
            self._emit("autoscroll-changed", self._autoscroll)

        self._requesting = None

    def _on_adj_value_changed(
        self, adj: Gtk.Adjustment, _pspec: GObject.ParamSpec
    ) -> None:
        if self._requesting is not None:
            return

        bottom = adj.get_upper() - adj.get_page_size()

        self._autoscroll = bottom - adj.get_value() < 1
        self._emit("autoscroll-changed", self._autoscroll)

        if self._upper_complete:
            self._request_history_at_upper = None
            if self._lower_complete:
                return

        if self._request_history_at_upper == adj.get_upper():
            # Abort here if we already did a history request and the upper
            # did not change. This can happen if we scroll very fast and the
            # value changes while the request has not been fulfilled.
            return

        self._request_history_at_upper = None

        distance = adj.get_page_size()
        if adj.get_value() < distance:
            # Load messages when we are near the top
            if self._upper_complete:
                return
            self._request_history_at_upper = adj.get_upper()
            # Workaround: https://gitlab.gnome.org/GNOME/gtk/merge_requests/395
            self.set_kinetic_scrolling(False)
            if not self._block_signals:
                self._emit("request-history", True)
            self._requesting = "before"

        elif adj.get_upper() - (adj.get_value() + adj.get_page_size()) < distance:
            # ..or near the bottom
            if self._lower_complete:
                return
            # Workaround: https://gitlab.gnome.org/GNOME/gtk/merge_requests/395
            self.set_kinetic_scrolling(False)
            if not self._block_signals:
                self._emit("request-history", False)
            self._requesting = "after"

    @property
    def contact(self) -> types.ChatContactT:
        assert self._contact is not None
        return self._contact

    def _get_row_at_index(self, index: int) -> BaseRow:
        row = self._list_box.get_row_at_index(index)
        assert row is not None
        return cast(BaseRow, row)

    def get_first_row(self) -> MessageRow | CallRow | FileTransferJingleRow | None:
        for row in iterate_listbox_children(self._list_box):
            if isinstance(row, MessageRow | CallRow | FileTransferJingleRow):
                return row
        return None

    def get_last_row(self) -> MessageRow | CallRow | FileTransferJingleRow | None:
        children = reversed(list(iterate_listbox_children(self._list_box)))
        for row in children:
            if isinstance(row, MessageRow | CallRow | FileTransferJingleRow):
                return row
        return None

    def get_first_message_row(self) -> MessageRow | None:
        for row in iterate_listbox_children(self._list_box):
            if isinstance(row, MessageRow):
                return row
        return None

    def get_last_message_row(self) -> MessageRow | None:
        children = reversed(list(iterate_listbox_children(self._list_box)))
        for row in children:
            if isinstance(row, MessageRow):
                return row
        return None

    def get_first_event_row(self) -> InfoMessage | MUCJoinLeft | None:
        for row in iterate_listbox_children(self._list_box):
            if isinstance(row, InfoMessage | MUCJoinLeft):
                return row
        return None

    def get_last_event_row(self) -> InfoMessage | MUCJoinLeft | None:
        children = reversed(list(iterate_listbox_children(self._list_box)))
        for row in children:
            if isinstance(row, InfoMessage | MUCJoinLeft):
                return row
        return None

    @staticmethod
    def _sort_func(row1: BaseRow, row2: BaseRow) -> int:
        if row1.timestamp == row2.timestamp:
            return 0
        return -1 if row1.timestamp < row2.timestamp else 1

    @staticmethod
    def _get_occupant_id(obj: mod.Message | DisplayedMarkerData) -> str | None:
        if isinstance(obj, DisplayedMarkerData):
            if obj.occupant is None:
                # Single chat, we don't store outgoing markers
                return str(ChatDirection.INCOMING)
            return obj.occupant.id

        if obj.type == MessageType.CHAT:
            # In single chat we use the direction to identify the user
            return str(obj.direction)

        if obj.occupant is None:
            return None
        return obj.occupant.id

    def _record_last_occupant_message(
        self, occupant_id: str, message: mod.Message
    ) -> None:
        last_message = self._last_occupant_messages.get(occupant_id)
        if last_message is None or last_message.timestamp < message.timestamp:
            self._last_occupant_messages[occupant_id] = message

    def _load_displayed_marker(self) -> None:
        assert self._contact is not None

        if isinstance(self._contact, GroupchatContact):
            markers = app.storage.archive.get_last_display_markers(
                self._contact.account, self._contact.jid
            )
        else:
            marker = app.storage.archive.get_last_display_marker(
                self._contact.account, self._contact.jid
            )
            if marker is None:
                return

            markers = [marker]

        for marker in markers:
            marker = DisplayedMarkerData.from_model(self._contact.account, marker)
            occupant_id = self._get_occupant_id(marker)
            if occupant_id is None:
                log.warning(
                    "Loaded unexpected marker from database, "
                    "unable to get occupant id: %s",
                    marker,
                )
                continue

            self._dm_occupant_markers[occupant_id] = marker
            self._dm_id_occupant_markers[marker.id][occupant_id] = marker

    def add_muc_subject(
        self, subject: MucSubject, timestamp: float | None = None
    ) -> None:
        muc_subject = MUCSubject(self.contact.account, subject, timestamp)
        self._insert_message(muc_subject)

    def add_muc_user_left(self, event: events.MUCUserLeft, error: bool = False) -> None:
        assert isinstance(self._contact, GroupchatContact)
        if not self._contact.settings.get("print_join_left"):
            return
        join_left = MUCJoinLeft(
            "muc-user-left",
            self._contact.account,
            event.nick,
            reason=event.reason,
            error=error,
            timestamp=event.timestamp,
        )
        self._insert_message(join_left)

    def add_muc_user_joined(self, event: events.MUCUserJoined) -> None:
        assert isinstance(self._contact, GroupchatContact)
        if not self._contact.settings.get("print_join_left"):
            return
        join_left = MUCJoinLeft(
            "muc-user-joined",
            self._contact.account,
            event.nick,
            timestamp=event.timestamp,
        )
        self._insert_message(join_left)

    def add_user_status(self, name: str, show: str, status: str) -> None:
        user_status = UserStatus(self.contact.account, name, show, status)
        self._insert_message(user_status)

    def add_info_message(self, text: str, timestamp: datetime | None = None) -> None:
        message = InfoMessage(self.contact.account, text, timestamp)
        self._insert_message(message)

    def add_file_transfer(self, transfer: HTTPFileTransfer) -> None:
        transfer_row = FileTransferRow(self.contact.account, transfer)
        transfer_row.connect("remove", self._remove_row)
        self._insert_message(transfer_row)

    def add_jingle_file_transfer(
        self,
        event: events.FileRequestReceivedEvent | events.FileRequestSent | None = None,
        message: Message | None = None,
    ) -> None:
        assert isinstance(self._contact, BareContact)
        jingle_transfer_row = FileTransferJingleRow(
            self._contact.account, self._contact, event=event, message=message
        )
        self._insert_message(jingle_transfer_row)

    def add_encryption_info(self, event: events.EncryptionInfo) -> None:
        assert self._contact is not None
        self._insert_message(EncryptionInfoRow(event))

    def add_call_message(
        self,
        event: events.JingleRequestReceived | None = None,
        message: Message | None = None,
    ) -> None:
        assert isinstance(self._contact, BareContact)
        call_row = CallRow(
            self._contact.account, self._contact, event=event, message=message
        )
        self._insert_message(call_row)

    def add_command_output(self, text: str, is_error: bool) -> None:
        command_output_row = CommandOutputRow(self.contact.account, text, is_error)
        self._insert_message(command_output_row)

    def add_message(self, message: mod.Message) -> None:
        message_row = MessageRow.from_db_row(self.contact, message)
        message_row.connect(
            "state-flags-changed", self._on_message_row_state_flags_changed
        )

        # Add all message ids and stanza ids of all message revisions
        for stanza_id in message.iter_stanza_ids():
            self._stanza_id_row_map[stanza_id] = message_row

        for message_id in message.iter_message_ids():
            self._message_id_row_map[message_id] = message_row

        self._insert_message(message_row)

        if occupant_id := self._get_occupant_id(message):
            self._record_last_occupant_message(occupant_id, message)
            self._update_displayed_markers(message)

    def _update_displayed_markers(self, message: mod.Message) -> None:
        if message.direction == ChatDirection.INCOMING:
            self._remove_obsolete_displayed_marker(message)

        displayed_id = message.get_displayed_id()
        if not displayed_id:
            return

        occupant_markers = self._dm_id_occupant_markers.get(displayed_id)
        if not occupant_markers:
            return

        if message.type in (MessageType.GROUPCHAT, MessageType.PM):
            if message.occupant is None:
                # Dont handle messages where we have no occupant-id
                return

            # Ignore markers from the sender for his own messages
            occupant_markers = occupant_markers.copy()
            occupant_markers.pop(message.occupant.id, None)

            markers = list(occupant_markers.values())
            if not markers:
                return

        else:
            marker = occupant_markers.get(str(ChatDirection.INCOMING))
            if marker is None:
                return

            if message.direction == ChatDirection.INCOMING:
                # Ignore markers from the sender for his own messages
                return

            markers = [marker]

        # Check if we have newer messages in the view as the marker points to
        for marker in list(markers):
            occupant_id = self._get_occupant_id(marker)
            if occupant_id is None:
                # Should be impossible
                log.warning("Unable to get occupant id: %s", marker)
                markers.remove(marker)
                continue

            if last_message := self._last_occupant_messages.get(occupant_id):
                if last_message.timestamp > message.timestamp:
                    markers.remove(marker)

        if not markers:
            return

        self._add_or_update_displayed_marker_row(message.timestamp, markers)

    def _remove_obsolete_displayed_marker(self, message: mod.Message) -> None:
        occupant_id = self._get_occupant_id(message)
        assert occupant_id is not None

        current_marker = self._dm_occupant_markers.get(occupant_id)
        if current_marker is None:
            return

        if isinstance(self._contact, GroupchatContact):
            message_row = self._stanza_id_row_map.get(current_marker.id)
        else:
            message_row = self._message_id_row_map.get(current_marker.id)

        if message_row is None:
            return

        # Check if current marker points to newer message in the view
        if message_row.timestamp > message.timestamp:
            return

        self._remove_displayed_marker(occupant_id)

    def update_displayed_markers(self, event: events.DisplayedReceived) -> None:
        marker = DisplayedMarkerData.from_model(event.account, event.marker)

        occupant_id = self._get_occupant_id(marker)
        if occupant_id is None:
            log.warning("Unable to get occupant id: %s", marker)
            return

        self._remove_displayed_marker(occupant_id)

        self._dm_occupant_markers[occupant_id] = marker
        self._dm_id_occupant_markers[marker.id][occupant_id] = marker

        if isinstance(self._contact, GroupchatContact):
            message_row = self._stanza_id_row_map.get(marker.id)
        else:
            message_row = self._message_id_row_map.get(marker.id)

        if message_row is None:
            return

        self._add_or_update_displayed_marker_row(message_row.timestamp, [marker])

    def _remove_displayed_marker(self, occupant_id: str) -> None:
        current_marker = self._dm_occupant_markers.get(occupant_id)
        if current_marker is None:
            return

        del self._dm_occupant_markers[occupant_id]
        del self._dm_id_occupant_markers[current_marker.id][occupant_id]

        row = self._dm_rows.get(current_marker.id)
        if row is None:
            return

        if occupant_id == str(ChatDirection.INCOMING):
            # Single Chat
            del self._dm_rows[current_marker.id]
            self._remove_row(row)

        else:
            row.remove_marker(current_marker)
            if not row.has_markers():
                del self._dm_rows[current_marker.id]
                self._remove_row(row)

    def _add_or_update_displayed_marker_row(
        self, timestamp: datetime, markers: list[DisplayedMarkerData]
    ) -> DisplayedRow | None:
        displayed_id = markers[0].id
        row = self._dm_rows.get(displayed_id)
        if row is not None:
            row.add_markers(markers)
            return None

        assert self._contact is not None
        row = DisplayedRow(self._contact.account, timestamp, markers)
        self._dm_rows[displayed_id] = row
        self._list_box.append(row)

    def _insert_message(self, message: BaseRow) -> None:
        self._list_box.append(message)

        self._add_date_row(message.timestamp)
        self._check_for_merge(message)

    def _add_date_row(self, timestamp: datetime) -> None:
        start_of_day = get_start_of_day(timestamp.astimezone())
        if start_of_day in self._active_date_rows:
            return

        date_row = DateRow(self.contact.account, start_of_day)
        self._active_date_rows.add(start_of_day)
        self._list_box.append(date_row)

        row = self._list_box.get_row_at_index(date_row.get_index() + 1)
        if row is None:
            return

        if not isinstance(row, MessageRow):
            return

        row.set_merged(False)

    def _check_for_merge(self, message: BaseRow) -> None:
        if not isinstance(message, MessageRow):
            return

        if not app.settings.get("chat_merge_consecutive_nickname"):
            return

        ancestor = self._find_ancestor(message)
        if ancestor is None:
            self._update_descendants(message)
        else:
            message.set_merged(message.is_mergeable(ancestor))

    def _find_ancestor(self, message: MessageRow) -> MessageRow | None:
        index = message.get_index()
        while index != 0:
            index -= 1
            row = self._list_box.get_row_at_index(index)
            if row is None:
                return None

            if isinstance(row, DisplayedRow):
                continue

            if not isinstance(row, MessageRow):
                return None

            if not message.is_same_sender(row):
                return None

            if not row.is_merged:
                return row
        return None

    def _update_descendants(self, message: MessageRow) -> None:
        index = message.get_index()
        while True:
            index += 1
            row = self._list_box.get_row_at_index(index)
            if row is None:
                return

            if isinstance(row, DisplayedRow):
                continue

            if not isinstance(row, MessageRow):
                return

            merge = message.is_mergeable(row)
            row.set_merged(merge)
            return

    def _on_message_row_state_flags_changed(
        self, row: MessageRow, previous_flags: Gtk.StateFlags
    ) -> None:
        if self._list_box.get_selection_mode() == Gtk.SelectionMode.MULTIPLE:
            # Message row selection active, hide MessageRowActions
            self._message_row_actions.hide_actions()
            return

        current_flags = self.get_state_flags()

        if (
            current_flags & Gtk.StateFlags.BACKDROP
            or previous_flags & Gtk.StateFlags.BACKDROP
        ):
            # Don't react to backdrop changes and focus changes when in background
            return

        if previous_flags & Gtk.StateFlags.PRELIGHT:
            self._message_row_actions.hide_actions()
        else:
            try:
                success, point = row.compute_point(self, Graphene.Point.zero())
            except GLib.Error:
                return

            if not success:
                return

            self._message_row_actions.update(point.y, row)

    def remove_message(self, pk: int) -> None:
        row = self.get_row_by_pk(pk)
        if row is None:
            return

        self._remove_from_maps(row)
        index = row.get_index()
        self._remove_row(row)
        decendant_row = self._list_box.get_row_at_index(index)
        if isinstance(decendant_row, MessageRow):
            # Unset possible merged state if we delete a 'top level' message.
            # Checks for same sender etc. are not necessary, since we simply
            # unset merged state.
            decendant_row.set_merged(False)

    def _remove_from_maps(self, row: MessageRow) -> None:
        for key, val in dict(self._message_id_row_map).items():
            if val is row:
                del self._message_id_row_map[key]

        for key, val in dict(self._stanza_id_row_map).items():
            if val is row:
                del self._stanza_id_row_map[key]

    def acknowledge_message(self, event: events.MessageAcknowledged) -> None:
        row = self.get_row_by_pk(event.pk)
        if row is None:
            return

        if event.stanza_id is not None:
            self._stanza_id_row_map[event.stanza_id] = row
        row.set_acknowledged(event.pk)
        self._check_for_merge(row)

    def scroll_to_message_and_highlight(self, pk: int) -> None:
        highlight_row = None
        for row in cast(list[BaseRow], iterate_listbox_children(self._list_box)):
            if row.pk == pk:
                highlight_row = row
                break

        if highlight_row is None:
            return

        # Scroll ListBox to row and highlight it
        coordinates = highlight_row.translate_coordinates(self._list_box, 0, 0)
        if coordinates is None:
            return

        _x_coord, y_coord = coordinates
        _mimimum_site, natural_size = highlight_row.get_preferred_size()
        adjustment = self._list_box.get_adjustment()
        assert adjustment is not None
        adjustment.set_value(
            y_coord - (adjustment.get_page_size() - natural_size.height) / 2
        )

        highlight_row.remove_css_class("conversation-row-highlight")
        highlight_row.add_css_class("conversation-row-highlight")

        GLib.timeout_add(1500, self._remove_highligh_class, highlight_row)

    def _remove_highligh_class(self, highlight_row: BaseRow) -> None:
        highlight_row.remove_css_class("conversation-row-highlight")

    def scroll_to_end(self) -> None:
        self._scroll_to_pos_idle(self.get_vadjustment(), -1)

    def _get_row_by_message_id(self, message_id: str) -> MessageRow | None:
        return self._message_id_row_map.get(message_id)

    def _get_row_by_stanza_id(self, stanza_id: str) -> MessageRow | None:
        return self._stanza_id_row_map.get(stanza_id)

    def _get_message_row_by_direction(
        self, pk: int, direction: Direction | None = None
    ) -> MessageRow | None:
        row = self.get_row_by_pk(pk)
        if row is None:
            return None

        if direction is None:
            return row

        index = row.get_index()
        while True:
            if direction == Direction.PREV:
                index -= 1
            else:
                index += 1

            row = self._list_box.get_row_at_index(index)
            if row is None:
                return None

            if isinstance(row, MessageRow):
                return row

    def get_prev_message_row(self, pk: int | None) -> MessageRow | None:
        if pk is None:
            return self.get_last_message_row()
        return self._get_message_row_by_direction(pk, direction=Direction.PREV)

    def get_next_message_row(self, pk: int | None) -> MessageRow | None:
        if pk is None:
            return None
        return self._get_message_row_by_direction(pk, direction=Direction.NEXT)

    def get_row_by_pk(self, pk: int) -> MessageRow | None:
        for row in cast(list[BaseRow], iterate_listbox_children(self._list_box)):
            if not isinstance(row, MessageRow):
                continue
            if pk in {row.pk, row.orig_pk}:
                return row

        return None

    def iter_rows(self) -> Generator[BaseRow, None, None]:
        yield from cast(list[BaseRow], iterate_listbox_children(self._list_box))

    def _remove_row(self, row: BaseRow) -> None:
        check_finalize(row)
        self._list_box.remove(row)

    def _remove_rows_by_type(self, row_type: str) -> None:
        for row in self.iter_rows():
            if row.type == row_type:
                self._remove_row(row)

    def update_call_rows(self) -> None:
        for row in cast(list[BaseRow], iterate_listbox_children(self._list_box)):
            if isinstance(row, CallRow):
                row.update()

    def correct_message(self, event: events.MessageCorrected) -> None:
        message_row = self._get_row_by_message_id(event.correction_id)
        if message_row is None:
            return

        if event.message.id is not None:
            self._message_id_row_map[event.message.id] = message_row

        if event.message.stanza_id is not None:
            self._stanza_id_row_map[event.message.stanza_id] = message_row

        message_row.update_corrections()

    def update_retractions(self, retraction_id: str) -> None:
        if isinstance(self._contact, GroupchatContact):
            message_row = self._get_row_by_stanza_id(retraction_id)
        else:
            message_row = self._get_row_by_message_id(retraction_id)

        if message_row is not None:
            message_row.update_retractions()

    def update_reactions(self, reaction_id: str) -> None:
        if isinstance(self._contact, GroupchatContact):
            message_row = self._get_row_by_stanza_id(reaction_id)
        else:
            message_row = self._get_row_by_message_id(reaction_id)

        if message_row is not None:
            message_row.update_reactions()

    def update_blocked_muc_users(self) -> None:
        assert self._client is not None
        assert self._contact is not None
        blocking_list = self._client.get_module("MucBlocking").get_blocking_list(
            self._contact.jid
        )

        for row in self.iter_rows():
            if not isinstance(row, MessageRow):
                continue
            if row.occupant_id is None:
                continue
            row.set_blocked(row.occupant_id in blocking_list)

    def set_receipt(self, id_: str) -> None:
        message_row = self._get_row_by_message_id(id_)
        if message_row is None:
            return

        message_row.set_receipt(id_)
        self._check_for_merge(message_row)

    def show_error(self, id_: str, error: StanzaError) -> None:
        message_row = self._get_row_by_message_id(id_)
        if message_row is not None:
            message_row.set_error(id_, to_user_string(error))

    def _on_contact_setting_changed(
        self, value: Any, setting: str, _account: str | None, _jid: JID | None
    ) -> None:
        if setting == "print_join_left":
            if value:
                return
            self._remove_rows_by_type("muc-user-joined")
            self._remove_rows_by_type("muc-user-left")

        if setting == "print_status":
            if value:
                return
            self._remove_rows_by_type("muc-user-status")



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import typing

import logging
from concurrent.futures import Future
from functools import partial
from pathlib import Path

from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk

from gajim.common import app
from gajim.common.const import IMAGE_MIME_TYPES
from gajim.common.const import VIDEO_MIME_TYPES
from gajim.common.helpers import load_file_async
from gajim.common.multiprocess.thumbnail import create_thumbnail
from gajim.common.multiprocess.video_thumbnail import (
    extract_video_thumbnail_and_properties,
)
from gajim.common.util.image import image_size
from gajim.common.util.image import is_image_animated

from gajim.gtk.preview.animated_image import AnimatedImage
from gajim.gtk.preview.animated_image_backend import AnimatedImageBackend
from gajim.gtk.preview.animated_image_fallback_backend import (
    AnimatedImageFallbackBackend,
)
from gajim.gtk.preview.file_control_buttons import FileControlButtons
from gajim.gtk.preview.misc import LoadingBox  # noqa: F401 # pyright: ignore
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import get_ui_string

log = logging.getLogger("gajim.gtk.preview.image")


@Gtk.Template.from_string(string=get_ui_string("preview/image.ui"))
class ImagePreviewWidget(Gtk.Box, SignalManager):
    __gtype_name__ = "ImagePreviewWidget"

    __gsignals__ = {
        "display-error": (
            GObject.SignalFlags.RUN_LAST | GObject.SignalFlags.ACTION,
            None,
            (),
        )
    }

    _stack: Gtk.Stack = Gtk.Template.Child()
    _content_overlay: Gtk.Overlay = Gtk.Template.Child()
    _image_button: Gtk.Button = Gtk.Template.Child()
    _picture: Gtk.Picture = Gtk.Template.Child()
    _play_image: Gtk.Image = Gtk.Template.Child()
    _file_control_buttons: FileControlButtons = Gtk.Template.Child()

    def __init__(
        self,
        filename: str,
        file_size: int,
        mime_type: str,
        orig_path: Path,
        thumb_path: Path,
    ) -> None:
        Gtk.Box.__init__(self)
        SignalManager.__init__(self)

        self._orig_path = orig_path
        self._thumb_path = thumb_path
        self._filename = filename
        self._mime_type = mime_type

        if mime_type in IMAGE_MIME_TYPES:
            self._type = "image"
        elif mime_type in VIDEO_MIME_TYPES:
            self._type = "video"
        else:
            raise ValueError("Not supported mime type: %s" % mime_type)

        content_hover_controller = Gtk.EventControllerMotion()
        self._connect(content_hover_controller, "enter", self._on_content_cursor_enter)
        self._connect(content_hover_controller, "leave", self._on_content_cursor_leave)
        self.add_controller(content_hover_controller)

        self._file_control_buttons.set_file_size(file_size)
        self._file_control_buttons.set_file_name(filename)
        self._file_control_buttons.set_path(orig_path)

        pointer_cursor = Gdk.Cursor.new_from_name("pointer")
        self._image_button.set_cursor(pointer_cursor)
        self._image_button.set_action_target_value(
            GLib.Variant("s", str(self._orig_path))
        )

        if not self._thumb_path.exists():
            self._create_thumbnail(self._type)
        else:
            load_file_async(self._thumb_path, self._on_thumb_load_finished)

    def _on_thumb_load_finished(
        self, data: bytes | None, error: GLib.Error | None, user_data: typing.Any
    ) -> None:
        if data is None:
            log.error("Loading thumbnail failed, %s: %s", self._thumb_path.name, error)
            self.emit("display-error")
            return

        self._thumbnail = data
        self._display_image_preview()

    def _create_thumbnail(self, type_: typing.Literal["image", "video"]) -> None:
        if type_ == "image":
            self._create_image_thumbnail()
        elif type_ == "video":
            self._create_video_thumbnail()

    def _create_image_thumbnail(self) -> None:
        assert self._thumb_path is not None
        assert self._orig_path is not None
        try:
            future = app.process_pool.submit(
                create_thumbnail,
                self._orig_path,
                self._thumb_path,
                app.settings.get("preview_size") * app.window.get_scale_factor(),
                self._mime_type,
            )
            future.add_done_callback(
                partial(GLib.idle_add, self._create_thumbnail_finished)
            )
        except Exception as error:
            log.warning("Creating thumbnail failed for: %s %s", self._orig_path, error)
            self.emit("display-error")

    def _create_video_thumbnail(self) -> None:
        assert self._orig_path is not None

        try:
            future = app.process_pool.submit(
                extract_video_thumbnail_and_properties,
                self._orig_path,
                self._thumb_path,
                app.settings.get("preview_size"),
            )
            future.add_done_callback(
                partial(GLib.idle_add, self._create_thumbnail_finished)
            )
        except Exception as error:
            log.warning("Creating thumbnail failed for: %s %s", self._orig_path, error)
            self.emit("display-error")

    def _create_thumbnail_finished(
        self, future: Future[tuple[bytes, dict[str, typing.Any]]]
    ) -> bool:
        try:
            thumbnail_bytes, _metadata = future.result()
        except Exception as error:
            log.exception(
                "Creating thumbnail failed for: %s %s", self._orig_path, error
            )
            self.emit("display-error")

        else:
            self._thumbnail = thumbnail_bytes
            self._display_image_preview()

        return GLib.SOURCE_REMOVE

    def _display_image_preview(self) -> None:
        if app.is_installed("GST"):
            if self._mime_type == "image/gif" and is_image_animated(self._orig_path):
                self._display_animated_image_preview(AnimatedImageBackend)
                return

            if self._mime_type in [
                "image/webp",
                "image/avif",
                "image/png",
            ] and is_image_animated(self._orig_path):
                self._display_animated_image_preview(AnimatedImageFallbackBackend)
                return

        self._display_static_image_preview()

    def _image_preview_dimension(
        self, image_width: int, image_height: int
    ) -> tuple[int, int]:
        max_preview_size = app.settings.get("preview_size")
        if image_width > max_preview_size or image_height > max_preview_size:
            # Scale down with or height to max_preview_size
            if image_width > image_height:
                width = max_preview_size
                height = int(max_preview_size / image_width * image_height)
            else:
                width = int(max_preview_size / image_height * image_width)
                height = max_preview_size
        else:
            width = image_width
            height = image_height

        return width, height

    def _display_static_image_preview(self) -> None:
        try:
            texture = Gdk.Texture.new_from_bytes(GLib.Bytes.new(self._thumbnail))
        except GLib.Error:
            log.exception("Could not load image %s", self._filename)
            self.emit("display-error")
            return

        texture_width = texture.get_width()
        texture_height = texture.get_height()

        width, height = self._image_preview_dimension(texture_width, texture_height)

        # Set minimum height of 100 to avoid button overlay covering entire image
        self._content_overlay.set_size_request(width, max(height, 100))
        self._image_button.set_tooltip_text(self._filename)
        self._picture.set_paintable(texture)

        if self._type == "video":
            self._image_button.add_css_class("preview-video-overlay")
            width, height = self._content_overlay.get_size_request()
            self._play_image.set_pixel_size(min(width, height) // 3)
            self._play_image.set_visible(True)

        self._stack.set_visible_child_name("preview")

    def _display_animated_image_preview(
        self, backend: typing.Any[AnimatedImageBackend, AnimatedImageFallbackBackend]
    ) -> None:
        image_width, image_height = image_size(self._orig_path)
        width, height = self._image_preview_dimension(image_width, image_height)

        animated_image = AnimatedImage(
            self._thumb_path, self._orig_path, width, height, backend
        )
        pointer_cursor = Gdk.Cursor.new_from_name("pointer")
        animated_image.set_cursor(pointer_cursor)

        self._connect(animated_image, "error", self._on_animated_image_error)

        self._content_overlay.set_size_request(width, height)
        self._content_overlay.set_child(animated_image)
        self._image_button.set_tooltip_text(self._filename)
        self._stack.set_visible_child_name("preview")

    def _on_animated_image_error(self, _animated_image: AnimatedImage) -> None:
        Gio.AppInfo.launch_default_for_uri(self._orig_path.as_uri(), None)

    def _on_content_cursor_enter(
        self,
        _controller: Gtk.EventControllerMotion,
        _x: int,
        _y: int,
    ) -> None:
        self._file_control_buttons.set_visible(True)

    def _on_content_cursor_leave(
        self,
        _controller: Gtk.EventControllerMotion,
    ) -> None:
        self._file_control_buttons.set_visible(False)

    def do_unroot(self) -> None:
        Gtk.Box.do_unroot(self)
        self._disconnect_all()



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any

from gi.repository import Adw
from gi.repository import Gdk
from gi.repository import Gtk

from gajim.common import app
from gajim.common import ged
from gajim.common import passwords
from gajim.common import types
from gajim.common.const import ClientState
from gajim.common.const import TLS_VERSION_STRINGS
from gajim.common.events import AccountDisabled
from gajim.common.events import AccountEnabled
from gajim.common.ged import EventHelper
from gajim.common.i18n import _
from gajim.common.i18n import p_

from gajim.gtk.alert import AlertDialog
from gajim.gtk.alert import CancelDialogResponse
from gajim.gtk.alert import DialogResponse
from gajim.gtk.certificate_dialog import CertificatePage
from gajim.gtk.const import Setting
from gajim.gtk.const import SettingKind
from gajim.gtk.const import SettingType
from gajim.gtk.omemo_trust_manager import OMEMOTrustManager
from gajim.gtk.preference.archiving import ArchivingPreferences
from gajim.gtk.preference.blocked_contacts import BlockedContacts
from gajim.gtk.preference.manage_roster import ManageRoster
from gajim.gtk.preference.widgets import CopyButton
from gajim.gtk.preference.widgets import PlaceholderBox
from gajim.gtk.settings import GajimPreferencePage
from gajim.gtk.settings import GajimPreferencesGroup
from gajim.gtk.settings import SignalManager
from gajim.gtk.settings import SpinRange
from gajim.gtk.structs import ExportHistoryParam
from gajim.gtk.util.misc import get_ui_string
from gajim.gtk.util.window import open_window


class AccountGeneralGroup(GajimPreferencesGroup):
    def __init__(self, account: str) -> None:
        GajimPreferencesGroup.__init__(
            self, key="account-general", account=account, title=_("General")
        )

        workspaces = self._get_workspaces()

        settings = [
            Setting(
                SettingKind.ENTRY,
                _("Label"),
                SettingType.ACCOUNT_CONFIG,
                "account_label",
                desc=_("Add a label to distinguish accounts in Gajim"),
            ),
            Setting(
                SettingKind.DROPDOWN,
                _("Default Workspace"),
                SettingType.ACCOUNT_CONFIG,
                "default_workspace",
                props={"data": workspaces},
                desc=_("Chats from this account will use this workspace by default"),
            ),
            Setting(
                SettingKind.COLOR,
                _("Color"),
                SettingType.ACCOUNT_CONFIG,
                "account_color",
                desc=_("Recognize your account by color"),
            ),
            Setting(
                SettingKind.SUBPAGE,
                _("Login"),
                SettingType.DIALOG,
                desc=_("Change your account’s password, etc."),
                bind="account::anonymous_auth",
                inverted=True,
                props={"subpage": f"{self.account}-general-login"},
            ),
            Setting(
                SettingKind.SWITCH,
                _("Connect on startup"),
                SettingType.ACCOUNT_CONFIG,
                "autoconnect",
            ),
            Setting(
                SettingKind.SWITCH,
                _("Global Status"),
                SettingType.ACCOUNT_CONFIG,
                "sync_with_global_status",
                desc=_("Synchronize the status of all accounts"),
            ),
            Setting(
                SettingKind.SWITCH,
                _("Remember Last Status"),
                SettingType.ACCOUNT_CONFIG,
                "restore_last_status",
                desc=_("Restore status and status message of your last session"),
            ),
            # Setting(
            #     SettingKind.SWITCH,
            #     _("Use file transfer proxies"),
            #     SettingType.ACCOUNT_CONFIG,
            #     "use_ft_proxies",
            # ),
        ]

        for setting in settings:
            self.add_setting(setting)

    @staticmethod
    def _get_workspaces() -> dict[str, str]:
        workspaces: dict[str, str] = {"": _("Disabled")}
        for workspace_id in app.settings.get_workspaces():
            name = app.settings.get_workspace_setting(workspace_id, "name")
            workspaces[workspace_id] = name
        return workspaces


class AccountGeneralAdvancedGroup(GajimPreferencesGroup):
    def __init__(self, account: str) -> None:
        GajimPreferencesGroup.__init__(
            self,
            key="account-general-advanced",
            account=account,
            description=_(
                "Removing the account will delete most related data. "
                "If you want to disconnect the account for a short time, "
                "use the status selector."
            ),
            title=_("Danger Zone"),
        )

        self.add(AccountActiveSwitch(account))

        settings = [
            Setting(
                SettingKind.GENERIC,
                _("Remove Account"),
                SettingType.VALUE,
                None,
                desc=_("This will remove the account and delete all data"),
                props={
                    "button-text": _("Remove…"),
                    "button-style": "destructive-action",
                    "button-callback": self._on_remove_account,
                },
            ),
        ]

        for setting in settings:
            self.add_setting(setting)

    def _on_remove_account(self, button: Gtk.Button) -> None:
        open_window("RemoveAccount", account=self.account)


class AccountPrivacyGroup(GajimPreferencesGroup):
    def __init__(self, account: str) -> None:
        GajimPreferencesGroup.__init__(
            self, key="account-privacy", account=account, title=_("Privacy")
        )

        self._client: types.Client | None = None
        if app.account_is_connected(account):
            self._client = app.get_client(account)

        history_max_age = {
            -1: _("Forever"),
            0: _("Until Gajim is Closed"),
            86400: _("1 Day"),
            604800: _("1 Week"),
            2629743: _("1 Month"),
            7889229: _("3 Months"),
            15778458: _("6 Months"),
            31556926: _("1 Year"),
        }

        chatstate_entries = {
            "disabled": _("Disabled"),
            "composing_only": _("Typing Only"),
            "all": _("All"),
        }

        encryption_entries = {
            "": _("Unencrypted"),
            "OMEMO": "OMEMO",
            "OpenPGP": "OpenPGP",
            "PGP": "PGP",
        }

        param = ExportHistoryParam(account=account, jid=None)

        settings = [
            Setting(
                SettingKind.DROPDOWN,
                _("Default Encryption"),
                SettingType.ACCOUNT_CONFIG,
                "encryption_default",
                desc=_(
                    "Encryption method to use unless overridden on a per-contact basis"
                ),
                props={"data": encryption_entries},
            ),
            Setting(
                SettingKind.SWITCH,
                _("Idle Time"),
                SettingType.ACCOUNT_CONFIG,
                "send_idle_time",
                callback=self._send_idle_time,
                desc=_("Disclose the time of your last activity"),
            ),
            Setting(
                SettingKind.SWITCH,
                _("Operating System"),
                SettingType.ACCOUNT_CONFIG,
                "send_os_info",
                callback=self._send_os_info,
                desc=_(
                    "Disclose information about the operating system you currently use"
                ),
            ),
            Setting(
                SettingKind.SWITCH,
                _("Media Playback"),
                SettingType.ACCOUNT_CONFIG,
                "publish_tune",
                callback=self._publish_tune,
                desc=_(
                    "Disclose information about media that is "
                    "currently being played on your system."
                ),
            ),
            Setting(
                SettingKind.SWITCH,
                _("Ignore Unknown Contacts"),
                SettingType.ACCOUNT_CONFIG,
                "ignore_unknown_contacts",
                desc=_("Ignore everything from contacts not in your contact list"),
            ),
            Setting(
                SettingKind.SWITCH,
                _("Send Message Receipts"),
                SettingType.ACCOUNT_CONFIG,
                "answer_receipts",
                desc=_("Tell your contacts if you received a message"),
            ),
            Setting(
                SettingKind.DROPDOWN,
                _("Send Chat Indicators"),
                SettingType.ACCOUNT_CONFIG,
                "send_chatstate_default",
                desc=_("Default for chats (e.g. Contact is typing)"),
                props={
                    "data": chatstate_entries,
                    "button-text": _("Reset"),
                    "button-tooltip": _("Reset all chats to the current default value"),
                    "button-style": "destructive-action",
                    "button-callback": self._reset_send_chatstate,
                },
            ),
            Setting(
                SettingKind.DROPDOWN,
                _("Send Chat Indicators in Group Chats"),
                SettingType.ACCOUNT_CONFIG,
                "gc_send_chatstate_default",
                desc=_("Default for group chats (e.g. Contact is typing)"),
                props={
                    "data": chatstate_entries,
                    "button-text": _("Reset"),
                    "button-tooltip": _(
                        "Reset all group chats to the current default value"
                    ),
                    "button-style": "destructive-action",
                    "button-callback": self._reset_gc_send_chatstate,
                },
            ),
            Setting(
                SettingKind.SWITCH,
                _("Read Receipts"),
                SettingType.VALUE,
                app.settings.get_account_setting(account, "send_marker_default"),
                callback=self._send_read_marker,
                desc=_("Default for chats and private group chats"),
                props={
                    "button-text": _("Reset"),
                    "button-tooltip": _("Reset all chats to the current default value"),
                    "button-style": "destructive-action",
                    "button-callback": self._reset_send_read_marker,
                },
            ),
            Setting(
                SettingKind.SWITCH,
                _("Update Timezone"),
                SettingType.ACCOUNT_CONFIG,
                "update_timezone",
                desc=_("Update your profile's timezone automatically"),
            ),
            Setting(
                SettingKind.SWITCH,
                _("Sync Group Chat Block List"),
                SettingType.ACCOUNT_CONFIG,
                "sync_muc_blocks",
                callback=self._sync_blocks,
                enabled_func=self._get_sync_blocks_enabled,
                desc=_("Synchronize group chat block list with other devices"),
            ),
            Setting(
                SettingKind.DROPDOWN,
                _("Keep Chat History"),
                SettingType.ACCOUNT_CONFIG,
                "chat_history_max_age",
                props={"data": history_max_age},
                desc=_("How long Gajim should keep your chat history"),
            ),
            Setting(
                SettingKind.ACTION,
                _("Export Chat History"),
                SettingType.ACTION,
                "app.export-history",
                props={"variant": param.to_variant()},
                desc=_("Export your chat history from Gajim"),
            ),
        ]

        for setting in settings:
            self.add_setting(setting)

    def _reset_send_chatstate(self, button: Gtk.Button) -> None:
        button.set_sensitive(False)
        app.settings.set_contact_settings("send_chatstate", self.account, None)

    def _reset_gc_send_chatstate(self, button: Gtk.Button) -> None:
        button.set_sensitive(False)
        app.settings.set_group_chat_settings("send_chatstate", self.account, None)

    def _send_idle_time(self, state: bool, _data: Any) -> None:
        if self._client is not None:
            self._client.get_module("LastActivity").set_enabled(state)

    def _send_os_info(self, state: bool, _data: Any) -> None:
        if self._client is not None:
            self._client.get_module("SoftwareVersion").set_enabled(state)

    def _publish_tune(self, state: bool, _data: Any) -> None:
        if self._client is not None:
            self._client.get_module("UserTune").set_enabled(state)

    def _send_read_marker(self, state: bool, _data: Any) -> None:
        assert self.account is not None
        app.settings.set_account_setting(self.account, "send_marker_default", state)
        app.settings.set_account_setting(
            self.account, "gc_send_marker_private_default", state
        )

    def _reset_send_read_marker(self, button: Gtk.Button) -> None:
        button.set_sensitive(False)
        app.settings.set_contact_settings("send_marker", self.account, None)
        app.settings.set_group_chat_settings(
            "send_marker", self.account, None, context="private"
        )

    def _get_sync_blocks_enabled(self) -> bool:
        if self._client is None:
            return False

        if not self._client.state.is_available:
            return False

        return self._client.get_module("Bookmarks").nativ_bookmarks_used

    def _sync_blocks(self, state: bool, _data: Any) -> None:
        if self._client is not None and state:
            self._client.get_module("MucBlocking").merge_blocks()


class AccountOmemoSettingsGroup(GajimPreferencesGroup):
    def __init__(self, account: str) -> None:
        GajimPreferencesGroup.__init__(
            self, key="account-omemo-settings", account=account
        )

        self.set_title(_("Trust Management"))
        wiki_url = "https://dev.gajim.org/gajim/gajim/-/wikis/help/OMEMO"
        link_text = _("Read more about blind trust")
        self.set_description(f'<a href="{wiki_url}">{link_text}</a>')

        settings = [
            Setting(
                SettingKind.SWITCH,
                _("Blind Trust"),
                SettingType.ACCOUNT_CONFIG,
                "omemo_blind_trust",
                desc=_("Blindly trust new devices until you verify them"),
            )
        ]

        for setting in settings:
            self.add_setting(setting)


class AccountOmemoTrustGroup(GajimPreferencesGroup):
    def __init__(self, account: str) -> None:
        GajimPreferencesGroup.__init__(self, key="account-omemo-trust", account=account)

        omemo_trust_manager = OMEMOTrustManager(account)
        self.add(omemo_trust_manager)


class AccountConnectionGroup(GajimPreferencesGroup):
    def __init__(self, account: str) -> None:
        GajimPreferencesGroup.__init__(
            self, key="account-connection", account=account, title=_("Connection")
        )

        settings = [
            Setting(
                SettingKind.SUBPAGE,
                _("Details"),
                SettingType.DIALOG,
                props={"subpage": f"{self.account}-connection-details"},
            ),
            Setting(
                SettingKind.SUBPAGE,
                _("Certificate"),
                SettingType.DIALOG,
                desc=_("Details about the certificate used by this connection"),
                props={"subpage": f"{self.account}-connection-certificate"},
            ),
            Setting(
                SettingKind.SUBPAGE,
                _("Hostname"),
                SettingType.ACCOUNT_CONFIG,
                "use_custom_host",
                desc=_("Manually set the hostname for the server"),
                props={"subpage": f"{self.account}-connection-hostname"},
            ),
            Setting(
                SettingKind.DROPDOWN,
                _("Proxy"),
                SettingType.ACCOUNT_CONFIG,
                "proxy",
                name="proxy",
                props={
                    "data": self._get_proxies(),
                    "button-icon-name": "lucide-settings-symbolic",
                    "button-callback": self._on_proxy_edit,
                },
            ),
            Setting(
                SettingKind.ENTRY,
                _("Resource"),
                SettingType.ACCOUNT_CONFIG,
                "resource",
                desc=_(
                    "Identifier for this device. "
                    "Don’t change this unless you know what you are doing."
                ),
            ),
            Setting(
                SettingKind.SWITCH,
                _("Use Unencrypted Connection"),
                SettingType.ACCOUNT_CONFIG,
                "use_plain_connection",
                desc=_("Use an unencrypted connection to the server"),
            ),
            Setting(
                SettingKind.SWITCH,
                _("Confirm Unencrypted Connection"),
                SettingType.ACCOUNT_CONFIG,
                "confirm_unencrypted_connection",
                desc=_("Ask before connecting unencrypted"),
            ),
        ]

        for setting in settings:
            self.add_setting(setting)

    @staticmethod
    def _get_proxies() -> dict[str, str]:
        proxies = {"": _("System")}
        proxies.update({proxy: proxy for proxy in app.settings.get_proxies()})
        proxies["no-proxy"] = _("No Proxy")
        return proxies

    @staticmethod
    def _on_proxy_edit(*args: Any) -> None:
        open_window("ManageProxies")


@Gtk.Template(string=get_ui_string("preference/connection_details.ui"))
class AccountConnectionDetailsGroup(Adw.PreferencesGroup, SignalManager):
    __gtype_name__ = "AccountConnectionDetailsGroup"

    _clipboard_button: CopyButton = Gtk.Template.Child()
    _domain: Adw.ActionRow = Gtk.Template.Child()
    _dns: Adw.ActionRow = Gtk.Template.Child()
    _ip_port: Adw.ActionRow = Gtk.Template.Child()
    _websocket: Adw.ActionRow = Gtk.Template.Child()
    _connection_type: Adw.ActionRow = Gtk.Template.Child()
    _tls_version: Adw.ActionRow = Gtk.Template.Child()
    _cipher_suite: Adw.ActionRow = Gtk.Template.Child()
    _proxy_type: Adw.ActionRow = Gtk.Template.Child()
    _proxy_host: Adw.ActionRow = Gtk.Template.Child()

    def __init__(self, account: str) -> None:
        Adw.PreferencesGroup.__init__(self)
        SignalManager.__init__(self)

        client = app.get_client(account)
        nbxmpp_client = client.connection
        address = nbxmpp_client.current_address

        assert address is not None
        assert address.domain is not None
        self._domain.set_subtitle(address.domain)

        visible = address.service is not None
        self._dns.set_visible(visible)
        self._dns.set_subtitle(address.service or "")

        visible = nbxmpp_client.remote_address is not None
        self._ip_port.set_visible(visible)
        self._ip_port.set_subtitle(nbxmpp_client.remote_address or "")

        visible = address.uri is not None
        self._websocket.set_visible(visible)
        self._websocket.set_subtitle(address.uri or "")

        self._connection_type.set_subtitle(address.type.value)
        if address.type.is_plain:
            self._connection_type.add_css_class("error")

        assert nbxmpp_client is not None
        tls_version = TLS_VERSION_STRINGS.get(nbxmpp_client.tls_version or -1)
        self._tls_version.set_subtitle(tls_version or _("Not available"))

        self._cipher_suite.set_subtitle(nbxmpp_client.ciphersuite or _("Not available"))

        # Connection proxy
        proxy = address.proxy
        if proxy is not None:
            self._proxy_type.set_subtitle(proxy.type)
            self._proxy_host.set_subtitle(proxy.host or "-")

        self._connect(
            self._clipboard_button, "clicked", self._on_clipboard_button_clicked
        )

    def _on_clipboard_button_clicked(self, _widget: Gtk.Button) -> None:
        string = f"Domain: {self._domain.get_subtitle()}\n"
        if self._dns.is_visible():
            string += f"DNS: {self._dns.get_subtitle()}\n"
        if self._ip_port.is_visible():
            string += f"IP/Port: {self._ip_port.get_subtitle()}\n"
        if self._websocket.is_visible():
            string += f"WebSocket URL: {self._websocket.get_subtitle()}\n"

        string += (
            f"Type: {self._connection_type.get_subtitle()}\n"
            f"TLS Version: {self._tls_version.get_subtitle()}\n"
            f"Cipher Suite: {self._cipher_suite.get_subtitle()}\n"
            f"Proxy Type: {self._proxy_type.get_subtitle()}\n"
            f"Proxy Host: {self._proxy_host.get_subtitle()}\n"
        )
        app.window.get_clipboard().set(string)

    def do_unroot(self) -> None:
        Adw.PreferencesGroup.do_unroot(self)
        self._disconnect_all()


class AccountManageRosterGroup(GajimPreferencesGroup):
    def __init__(self, account: str) -> None:
        GajimPreferencesGroup.__init__(
            self,
            key="account-manage-roster",
            account=account,
            title=_("Contact List"),
        )

        self.add(ManageRoster(account))


class AccountBlockedContactsGroup(GajimPreferencesGroup):
    def __init__(self, account: str) -> None:
        GajimPreferencesGroup.__init__(
            self,
            key="account-blocked-contacts",
            account=account,
            title=_("Blocked Contacts"),
        )

        self.add(BlockedContacts(account))


class AccountArchivingGroup(GajimPreferencesGroup):
    def __init__(self, account: str) -> None:
        GajimPreferencesGroup.__init__(
            self, key="account-archiving", account=account, title=_("Archive")
        )

        settings = [
            Setting(
                SettingKind.GENERIC,
                _("Synchronize History…"),
                SettingType.VALUE,
                None,
                props={
                    "button-text": _("Open"),
                    "button-callback": self._on_snyc,
                },
            ),
        ]

        for setting in settings:
            self.add_setting(setting)

    def _on_snyc(self, _button: Gtk.Button) -> None:
        open_window("HistorySyncAssistant", account=self.account)


class AccountArchivingPreferencesGroup(GajimPreferencesGroup):
    def __init__(self, account: str) -> None:
        GajimPreferencesGroup.__init__(
            self,
            key="account-archiving-preferences",
            account=account,
            description=_(
                "Configure which messages will be archived on the server. "
                "Not archiving messages on the server may prevent them from "
                "synchronizing between your devices."
            ),
            title=_("Preferences"),
        )

        self.add(ArchivingPreferences(account))


class AccountAdvancedGroup(GajimPreferencesGroup):
    def __init__(self, account: str) -> None:
        GajimPreferencesGroup.__init__(
            self, key="account-advanced", account=account, title=_("Advanced")
        )

        settings = [
            Setting(
                SettingKind.SWITCH,
                _("Contact Information"),
                SettingType.ACCOUNT_CONFIG,
                "request_user_data",
                desc=_("Request contact information (Tune, Location)"),
            ),
            Setting(
                SettingKind.SWITCH,
                _("Accept all Contact Requests"),
                SettingType.ACCOUNT_CONFIG,
                "autoauth",
                desc=_("Automatically accept all contact requests"),
            ),
            # TODO Jingle FT
            # Setting(SettingKind.DROPDOWN,
            #         _('Filetransfer Preference'),
            #         SettingType.ACCOUNT_CONFIG,
            #         'filetransfer_preference',
            #         props={'data': {'httpupload': _('Upload Files'),
            #                         'jingle': _('Send Files Directly')}},
            #         desc=_('Preferred file transfer mechanism for '
            #                'file drag&drop on a chat window')),
            Setting(
                SettingKind.SWITCH,
                _("Security Labels"),
                SettingType.ACCOUNT_CONFIG,
                "enable_security_labels",
                desc=_(
                    "Show labels describing confidentiality of "
                    "messages, if the server supports XEP-0258"
                ),
            ),
            Setting(
                SettingKind.SWITCH,
                _("Synchronize joined group chats"),
                SettingType.ACCOUNT_CONFIG,
                "autojoin_sync",
                desc=_("Synchronize joined group chats with other devices."),
            ),
        ]

        for setting in settings:
            self.add_setting(setting)


class LoginGroup(GajimPreferencesGroup):
    def __init__(self, account: str) -> None:
        GajimPreferencesGroup.__init__(
            self, key="login", account=account, title=_("Login")
        )

        settings = [
            Setting(
                SettingKind.SWITCH,
                _("Save Password"),
                SettingType.ACCOUNT_CONFIG,
                "savepass",
                enabled_func=(
                    lambda: (
                        not app.settings.get("use_keyring")
                        or passwords.is_keyring_available()
                    )
                ),
                callback=self._on_save_password,
            ),
            Setting(
                SettingKind.DIALOG,
                _("Change Password"),
                SettingType.DIALOG,
                enabled_func=(lambda: app.account_is_connected(account)),
                props={"dialog": "ChangePassword"},
            ),
            Setting(
                SettingKind.SWITCH,
                _("Use GSSAPI"),
                SettingType.ACCOUNT_CONFIG,
                "enable_gssapi",
            ),
        ]

        for setting in settings:
            self.add_setting(setting)

    def _on_save_password(self, state: bool, _data: Any) -> None:
        if not state:
            assert self.account is not None
            passwords.delete_password(self.account)


class HostnameGroup(GajimPreferencesGroup):
    def __init__(self, account: str) -> None:
        GajimPreferencesGroup.__init__(
            self, key="hostname", account=account, title=_("Hostname")
        )

        type_values = ["START TLS", "DIRECT TLS", "PLAIN"]

        settings = [
            Setting(
                SettingKind.SWITCH,
                _("Enable"),
                SettingType.ACCOUNT_CONFIG,
                "use_custom_host",
            ),
            Setting(
                SettingKind.ENTRY,
                _("Hostname"),
                SettingType.ACCOUNT_CONFIG,
                "custom_host",
                bind="account::use_custom_host",
            ),
            Setting(
                SettingKind.SPIN,
                _("Port"),
                SettingType.ACCOUNT_CONFIG,
                "custom_port",
                bind="account::use_custom_host",
                props={"range_": SpinRange(0, 65535, 1, 0)},
            ),
            Setting(
                SettingKind.DROPDOWN,
                _("Type"),
                SettingType.ACCOUNT_CONFIG,
                "custom_type",
                bind="account::use_custom_host",
                props={"data": type_values},
            ),
        ]

        for setting in settings:
            self.add_setting(setting)


class AccountGeneralPage(GajimPreferencePage):
    key = "general"
    icon_name = "lucide-user-symbolic"
    label = _("General")

    def __init__(self, account: str) -> None:
        GajimPreferencePage.__init__(
            self,
            title=_("General – %(account)s") % {"account": account},
            groups=[],
            tag_prefix=f"{account}-",
        )

        self.add(AccountGeneralGroup(account))
        self.add(AccountGeneralAdvancedGroup(account))


class AccountPrivacyPage(GajimPreferencePage):
    key = "privacy"
    icon_name = "lucide-eye-symbolic"
    label = _("Privacy")

    def __init__(self, account: str) -> None:
        GajimPreferencePage.__init__(
            self,
            title=_("Privacy – %(account)s") % {"account": account},
            groups=[],
            tag_prefix=f"{account}-",
        )

        self.add(AccountPrivacyGroup(account))


class AccountOmemoPage(GajimPreferencePage):
    key = "encryption-omemo"
    icon_name = "lucide-lock-symbolic"
    label = _("Encryption (OMEMO)")

    def __init__(self, account: str) -> None:
        GajimPreferencePage.__init__(
            self,
            title=_("Encryption (OMEMO) – %(account)s") % {"account": account},
            groups=[],
            tag_prefix=f"{account}-",
        )

        self.add(AccountOmemoSettingsGroup(account))
        self.add(AccountOmemoTrustGroup(account))


class AccountConnectionDetailsPage(GajimPreferencePage):
    key = "connection-details"

    def __init__(self, account: str) -> None:
        GajimPreferencePage.__init__(
            self,
            title=_("Details - %(account)s") % {"account": account},
            groups=[],
            tag_prefix=f"{account}-",
        )

        if not app.account_is_available(account):
            self.set_content(PlaceholderBox(valign=Gtk.Align.CENTER))
            return

        self.add(AccountConnectionDetailsGroup(account))


class AccountConnectionCertificatePage(GajimPreferencePage):
    key = "connection-certificate"

    def __init__(self, account: str) -> None:
        GajimPreferencePage.__init__(
            self,
            title=_("Certificate - %(account)s") % {"account": account},
            groups=[],
            tag_prefix=f"{account}-",
        )

        if not app.account_is_available(account):
            self.set_content(PlaceholderBox(valign=Gtk.Align.CENTER))
            return

        client = app.get_client(account)
        if client.certificate is None:
            self.set_content(
                Gtk.Label(label=_("No certificate available"), valign=Gtk.Align.CENTER)
            )
            return

        self.set_content(CertificatePage(account, client.certificate))


class AccountConnectionPage(GajimPreferencePage):
    key = "connection"
    icon_name = "lucide-globe-symbolic"
    label = _("Connection")

    def __init__(self, account: str) -> None:
        GajimPreferencePage.__init__(
            self,
            title=_("Connection – %(account)s") % {"account": account},
            groups=[],
            tag_prefix=f"{account}-",
        )

        self.add(AccountConnectionGroup(account))


class AccountManageRosterPage(GajimPreferencePage):
    key = "manage-roster"
    icon_name = "lucide-users-symbolic"
    label = _("Contact List")

    def __init__(self, account: str) -> None:
        GajimPreferencePage.__init__(
            self,
            title=_("Contact List – %(account)s") % {"account": account},
            groups=[],
            tag_prefix=f"{account}-",
        )

        if not app.account_is_available(account):
            self.set_content(PlaceholderBox(valign=Gtk.Align.CENTER))
            return

        self.add(AccountManageRosterGroup(account))


class AccountBlockedContactsPage(GajimPreferencePage):
    key = "blocked-contacts"
    icon_name = "lucide-user-round-x-symbolic"
    label = _("Blocked Contacts")

    def __init__(self, account: str) -> None:
        GajimPreferencePage.__init__(
            self,
            title=_("Blocked Contacts – %(account)s") % {"account": account},
            groups=[],
            tag_prefix=f"{account}-",
        )

        if not app.account_is_available(account):
            self.set_content(PlaceholderBox(valign=Gtk.Align.CENTER))
            return

        self.add(AccountBlockedContactsGroup(account))


class AccountArchivingPage(GajimPreferencePage):
    key = "archiving"
    icon_name = "lucide-database-backup-symbolic"
    label = _("Archiving")

    def __init__(self, account: str) -> None:
        GajimPreferencePage.__init__(
            self,
            title=_("Archiving – %(account)s") % {"account": account},
            groups=[],
            tag_prefix=f"{account}-",
        )

        if not app.account_is_available(account):
            self.set_content(PlaceholderBox(valign=Gtk.Align.CENTER))
            return

        self.add(AccountArchivingGroup(account))
        self.add(AccountArchivingPreferencesGroup(account))


class LoginPage(GajimPreferencePage):
    key = "general-login"

    def __init__(self, account: str) -> None:
        GajimPreferencePage.__init__(
            self,
            title=_("Login – %(account)s") % {"account": account},
            groups=[],
            tag_prefix=f"{account}-",
        )

        self.add(LoginGroup(account))


class HostnamePage(GajimPreferencePage):
    key = "connection-hostname"

    def __init__(self, account: str) -> None:
        GajimPreferencePage.__init__(
            self,
            title=_("Hostname – %(account)s") % {"account": account},
            groups=[],
            tag_prefix=f"{account}-",
        )

        self.add(HostnameGroup(account))


class AccountAdvancedPage(GajimPreferencePage):
    key = "advanced"
    icon_name = "lucide-settings-symbolic"
    label = _("Advanced")

    def __init__(self, account: str) -> None:
        GajimPreferencePage.__init__(
            self,
            title=_("Advanced – %(account)s") % {"account": account},
            groups=[],
            tag_prefix=f"{account}-",
        )

        self.add(AccountAdvancedGroup(account))


class AccountActiveSwitch(Adw.ActionRow, SignalManager, EventHelper):
    def __init__(self, account: str) -> None:
        Adw.ActionRow.__init__(self, title=_("Enable Account"))
        SignalManager.__init__(self)
        EventHelper.__init__(self)

        self._account = account

        active = app.settings.get_account_setting(account, "active")

        self._label = Gtk.Label(margin_end=12)
        self.add_suffix(self._label)

        self._switch = Gtk.Switch(active=active, valign=Gtk.Align.CENTER)
        self.add_suffix(self._switch)
        self.set_activatable_widget(self._switch)
        self._update_label()

        self.register_events(
            [
                ("account-enabled", ged.GUI2, self._on_account_state_changed),
                ("account-disabled", ged.GUI2, self._on_account_state_changed),
            ]
        )

        self._connect(self._switch, "state-set", self._on_state_set, account)

    def do_unroot(self) -> None:
        Adw.ActionRow.do_unroot(self)
        self._disconnect_all()
        self.unregister_events()
        app.check_finalize(self)

    def _update_label(self) -> None:
        if self._switch.get_active():
            self._label.set_text(p_("Switch", "On"))
        else:
            self._label.set_text(p_("Switch", "Off"))

    def _on_account_state_changed(
        self, event: AccountEnabled | AccountDisabled
    ) -> None:
        if event.account != self._account:
            return

        state = isinstance(event, AccountEnabled)
        self._switch.set_state(state)
        self._update_label()

    def _on_state_changed(
        self, client: types.Client, _signal_name: str, client_state: ClientState
    ) -> None:
        if client_state.is_disconnected:
            app.app.disable_account(client.account)

    def _on_state_set(self, switch: Gtk.Switch, state: bool, account: str) -> int:
        def _on_response(response_id: str) -> None:
            if response_id == "disable":
                client = app.get_client(account)
                client.connect_signal("state-changed", self._on_state_changed)
                client.change_status("offline", "offline")

            switch.set_state(state)

        account_is_active = app.settings.get_account_setting(account, "active")
        if account_is_active == state:
            return Gdk.EVENT_PROPAGATE

        if account_is_active and not app.get_client(account).state.is_disconnected:
            account_label = app.get_account_label(account)
            AlertDialog(
                _("Disable Account?"),
                _(
                    "Account %(name)s is still connected\n"
                    "All chat and group chat windows will be closed."
                )
                % {"name": account_label},
                responses=[
                    CancelDialogResponse(),
                    DialogResponse(
                        "disable", _("_Disable Account"), appearance="destructive"
                    ),
                ],
                callback=_on_response,
            )
            return Gdk.EVENT_STOP

        if state:
            app.app.enable_account(account)
        else:
            app.app.disable_account(account)

        return Gdk.EVENT_PROPAGATE
